"""
Contains the base Exposure class.
"""

import datetime
import os
import threading
import time
import ast
from typing import Union, List, Optional

import numpy

import azcam
import azcam.exceptions
import azcam.utils
from azcam.tools.exposure_filename import Filename
from azcam.tools.exposure_sendimage import SendImage
from azcam.tools.exposure_obstime import ObsTime
from azcam.header import Header, ObjectHeaderMethods
from azcam.image import Image
from azcam.tools.tools import Tools


class Exposure(Tools, Filename, ObjectHeaderMethods):
    """
    The base exposure tool.
    Usually implemented as the "exposure" tool.
    Only required attributes and stub methods are defined here. Additional
    methods and attributes are added as needed in exposure-specific classes
    which should inherit this class.
    """

    def __init__(self, tool_id="exposure", description=None):
        Tools.__init__(self, tool_id, description)
        Filename.__init__(self)

        self.sendimage = SendImage()
        self.obstime = ObsTime()
        self.image = Image()

        # exposure flags, may be used anywhere
        self.exposureflags = {
            "NONE": 0,
            "EXPOSING": 1,
            "ABORT": 2,
            "PAUSE": 3,
            "RESUME": 4,
            "READ": 5,
            "PAUSED": 6,
            "READOUT": 7,
            "SETUP": 8,
            "WRITING": 9,
            "GUIDEERROR": 10,
            "ERROR": 11,
        }

        azcam.db.exposureflags = self.exposureflags

        self.exposureflags_rev = {v: k for k, v in self.exposureflags.items()}

        # exposure flag defining state of current exposure
        self.exposure_flag = self.exposureflags["NONE"]

        # current image type, 'zero', 'object', 'dark', 'flat', 'ramp', etc
        self.image_type = "zero"
        # default imagetypes
        self.image_types = ["zero", "object", "flat", "dark"]
        # dictionary of shutter states for imagetypes {imagetype:ShutterState}
        self.shutter_dict = {
            "zero": 0,
            "object": 1,
            "flat": 1,
            "dark": 0,
            "ramp": 1,
        }
        # True to flush detector before exposures
        self.flush_array = 1
        # True to display an image after readout
        self.display_image = 1
        # True to send image to remote image server after readout
        self.send_image = 0

        self.message = ""  # exposure status message
        self.guide_status = 0
        self.guide_image_copy = 0

        # TdiMode flag, 0=not in TDI mode, 1=TDI mode
        self.tdi_mode = 0
        # TdiDelay mode
        self.tdi_delay = 5
        # ParDelay mode
        self.par_delay = 5

        # guide mode
        self.guide_mode = 0

        # True when exposure type is a comparision, to turn on comp lamps
        self.comp_exposure = 0
        # True when in a comparision exposure sequence so lamps stay on
        self.comp_sequence = 0

        # True when exposure is completed, then toggled off
        self.completed = 0

        # requested exposure time in seconds
        self.exposure_time = 1.0
        # remaining exposure time in seconds for an exposure in progress
        self.exposure_time_remaining = 0.0
        # actual elapsed exposure time in seconds of last/current exposure
        self.exposure_time_actual = 0.0
        # exposure time saved for each exposure, used for zeros
        self.exposure_time_saved = 0.0
        # total time in seconds an exposure was paused
        self.paused_time = 0.0
        # starting clock paused time of exposure
        self.paused_time_start = 0.0
        # actual elapsed dark time of last/current exposure
        self.dark_time = 0.0
        # starting clock dark time of exposure
        self.dark_time_start = 0.0

        # True when in an exposure sequence
        self.is_exposure_sequence = 0
        # current exposure sequence number
        self.exposure_sequence_number = 1
        # total number of exposures in sequence
        self.exposure_sequence_total = 1
        # delay between sequence exposures in seconds
        self.exposure_sequence_delay = 0.0
        # sequence flush flag: -1=> use FlushArray, 0==> flush all, 1=> flush only first exposure, 2=> no flush
        self.exposure_sequence_flush = 0

        # remaining number of pixels to read for an exposure in progress
        self.pixels_remaining = 0

        # True to update headers in a thread to save time
        self.update_headers_in_background = 0
        self.updating_header = 0

        # True to save image file after exposure
        self.save_file = 1

        # file types
        self.filetypes = {"FITS": 0, "MEF": 1, "BIN": 2, "ASM": 6}
        self.filetype = self.filetypes["MEF"]

        # Exposure title
        self.title = ""

        # True to make image title the same as image type
        self.auto_title = 0

        # deinterlace mode; 1 = generic mode; x = ODI mode
        self.deinterlace_mode = 1

        # temporary image files
        self.temp_image_file = ""

        # filename of current image
        self.last_filename = ""

        # write data asynchronously
        self.write_async = 0

        # create the exposure header
        self.header = Header("Exposure")

        # flag indicating ROI has been changed
        self.new_roi = 0
        self.header.set_header("exposure", 1)

        # data order
        self.data_order = []

        self.imageheaderfile = ""

        self.pgress = 0  # debug

    def initialize(self):
        """
        Initialize exposure.
        """

        # call initialize() method on other tools
        for tool in azcam.db.tools_init:
            azcam.db.tools[tool].initialize()

        self.is_initialized = 1

        return

    def reset(self):
        """
        Reset exposure tool.
        """

        # initialize only once
        if not self.is_initialized:
            self.initialize()

        # set temporary filenames
        self.set_temp_files()

        # setup for exposures
        self.is_exposure_sequence = 0
        self.exposure_sequence_number = 1
        self.set_auto_title()
        azcam.db.abortflag = 0
        self.save_file = 1
        self.exposure_flag = self.exposureflags["NONE"]

        # call reset() method on other tools
        for tool in azcam.db.tools_reset:
            try:
                azcam.db.tools[tool].reset()
            except Exception as e:
                raise azcam.exceptions.AzcamError(f"Could not reset {tool}: {e}")

        return

    # **********************************************************************************************
    # Exposure control
    # **********************************************************************************************

    def start(self):
        """
        Allow custom operations at start of exposure.
        """

        try:
            azcam.db.tools["instrument"].exposure_start()
        except KeyError:
            pass
        try:
            azcam.db.tools["telescope"].exposure_start()
        except KeyError:
            pass

        return

    def finish(self):
        """
        Allow custom operations at end of exposure.
        """

        try:
            azcam.db.tools["instrument"].exposure_finish()
        except KeyError:
            pass
        try:
            azcam.db.tools["telescope"].exposure_finish()
        except KeyError:
            pass

        return

    def finished(self):
        if self.completed:
            return 1
        else:
            return 0

    def get_exposureflag(self):
        return [self.exposure_flag, self.exposureflags_rev[self.exposure_flag]]

    def test(self, exposure_time=0.0, shutter=0):
        """
        Make a test exposure.
        exposure_time is the exposure time in seconds
        shutter is 0 for closed and 1 for open
        title is the image title.
        """

        old_testimage = self.test_image
        old_imagetype = self.image_type
        old_exposuretime = self.exposure_time
        self.test_image = 1

        if shutter:
            shutter_state = "object"
        else:
            shutter_state = "dark"

        self.expose(exposure_time, shutter_state, "test image")

        self.test_image = old_testimage
        self.image_type = old_imagetype
        self.exposure_time = old_exposuretime

        return

    def expose(self, exposure_time=-1, imagetype="", title=""):
        """
        Make a complete exposure.
        exposure_time is the exposure time in seconds
        imagetype is the type of exposure ('zero', 'object', 'flat', ...)
        title is the image title.
        """

        azcam.db.abortflag = 0

        imagetype = azcam.utils.dequote(imagetype)
        title = azcam.utils.dequote(title)

        # allow custom operations
        self.start()

        azcam.log("Exposure started")

        # if last exposure was aborted, warn before clearing flag
        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.exceptions.warning("Previous exposure was aborted")

        # begin
        if self.exposure_flag != self.exposureflags["ABORT"]:
            self.begin(exposure_time, imagetype, title)

        # integrate
        if self.exposure_flag != self.exposureflags["ABORT"]:
            self.integrate()

        # readout
        if (
            self.exposure_flag != self.exposureflags["ABORT"]
            and self.exposure_flag == self.exposureflags["READ"]
        ):
            try:
                self.readout()
            except azcam.exceptions.AzcamError:
                pass
        # end
        if self.exposure_flag != self.exposureflags["ABORT"]:
            self.end()

        self.exposure_flag = self.exposureflags["NONE"]
        self.completed = 1
        azcam.log("Exposure finished")

        # allow custom operations
        self.finish()

        return

    def expose1(
        self, exposure_time: float = -1, image_type: str = "", image_title: str = ""
    ):
        """
        Make a complete exposure with immediate return to caller.

        :param exposure_time: the exposure time in seconds
        :param image_type: type of exposure ('zero', 'object', 'flat', ...)
        :param image_title: image title, usually surrounded by double quotes
        """

        arglist = [exposure_time, image_type, image_title]
        thread = threading.Thread(target=self.expose, name="expose1", args=arglist)
        thread.start()

        return

    def guide(self, number_exposures=1):
        """
        Make a complete guider exposure sequence.
        NumberExposures is the number of exposures to make, -1 loop forever
        """

        AbortFlag = 0

        number_exposures = int(number_exposures)

        # system must be reset once before an exposure can be made
        if not azcam.db.tools["controller"].is_reset:
            azcam.db.tools["controller"].reset()

        # parameters for faster operation
        flusharray = self.flush_array
        azcam.log("Guide started")

        # this loop continues even for errors since data is sent to a seperate client receiving images
        LoopCount = 0
        while True:
            if 0:
                self.begin(exposure_time=-1, imagetype="object", title="guide image")

                # integrate
                self.integrate()

                # readout
                if self.exposure_flag == self.exposureflags["READ"]:
                    try:
                        self.readout()
                        self.guide_status = 1  # image read OK
                        self.guide_image_copy = self.image
                    except Exception:
                        self.guide_status = (
                            2  # image not read OK, but don't stop guide loop
                        )
                        self.image = self.guide_image_copy

                # image writing
                self.end()
                self.exposure_flag = self.exposureflags["NONE"]
            else:
                self.expose(-1, "object", "guide image")

            AbortFlag = azcam.db.abortflag
            if AbortFlag:
                break

            if number_exposures == -1:
                continue
            else:
                LoopCount += 1

            if LoopCount >= number_exposures:
                break

        # finish
        self.guide_status = 0
        self.flush_array = flusharray

        if AbortFlag:
            azcam.exceptions.warning("Guide aborted")
        else:
            azcam.log("Guide finished")

        return

    def guide1(self, number_exposures=1):
        """
        Make a complete guider exposure with an immediate return.
        NumberExposures is the number of exposures to make, -1 loop forever
        """

        arglist = [number_exposures]
        thread = threading.Thread(target=self.guide, name="guide1", args=arglist)
        thread.start()

        return

    def begin(self, exposure_time=-1, imagetype="", title=""):
        """
        Initiates the first part of an exposure, through image flushing.
        exposure_time is in seconds.
        imagetype is one of zero, object, flat, dark, ramp, ...
        """

        # system must be reset once before an exposure can be made
        x = self.is_exposure_sequence  # save this flag which is lost by reset
        if not azcam.db.tools["controller"].is_reset:
            self.reset()
            self.is_exposure_sequence = x

        # set exposure flag
        self.exposure_flag = self.exposureflags["SETUP"]

        # reset flags as new data coming
        self.image.valid = 0
        self.image.is_written = 0
        self.image.toggle = 0
        self.image.assembled = 0

        # clear the image header
        self.image.header.delete_all_items()
        self.image.header.delete_all_keywords()

        # update image size
        try:
            self.set_roi()  # bug?
        except Exception:
            pass

        if self.new_roi:
            self.image.data = numpy.empty(
                shape=[
                    self.image.focalplane.numamps_image,
                    self.image.focalplane.numpix_amp,
                ],
                dtype="<u2",
            )
            self.new_roi = 0

        # imagetype
        if imagetype != "":
            self.image_type = imagetype
        self.image.image_type = self.image_type
        imagetype = self.image_type.lower()

        # exposure times
        exposure_time = float(exposure_time)
        if exposure_time >= 0:
            self.exposure_time = exposure_time
        self.exposure_time_saved = self.exposure_time
        if imagetype == "zero":
            self.exposure_time = 0.0
        self.paused_time = 0.0  # reset paused time
        self.paused_time_start = 0.0  # reset paused start time
        self.exposure_time_remaining = self.exposure_time

        # update title and set OBJECT keyword using title
        self.set_image_title(title)

        # send exposure time to controller
        self.set_exposuretime(self.exposure_time)

        # set total number of pixels to readout
        self.pixels_remaining = self.image.focalplane.numpix_image

        # set CompExposure flag for any undefined image types (comp names)
        if imagetype not in self.image_types:
            self.comp_exposure = 1
        else:
            self.comp_exposure = 0

        if not self.guide_mode:  # for speed
            try:
                shutterstate = self.shutter_dict[imagetype]
            except KeyError:
                shutterstate = 1  # other types are comps, so open shutter
            azcam.db.tools["controller"].set_shutter_state(shutterstate)

        self.delete_keyword("COMPLAMP")

        # set comp lamps, turn on, set keyword
        if self.comp_exposure and azcam.db.tools["instrument"].is_enabled:
            if self.comp_sequence:  # lamps already on
                pass
            else:
                azcam.db.tools["instrument"].set_comps(imagetype)
                if not azcam.db.tools["instrument"].shutter_strobe:
                    azcam.db.tools["instrument"].comps_on()
            lampnames = " ".join(azcam.db.tools["instrument"].get_comps())
            self.set_keyword("COMPLAMP", lampnames, "Comp lamp names", "str")
            self.set_keyword("IMAGETYP", "comp", "Image type", "str")
            azcam.db.tools["instrument"].comps_delay()  # delay for lamp warmup
        else:
            if not self.guide_mode:
                try:
                    if (azcam.db.tools["instrument"] is not None) and azcam.db.tools[
                        "instrument"
                    ].is_enabled:
                        azcam.db.tools["instrument"].set_comps()  # reset
                except KeyError:
                    pass
                self.set_keyword("IMAGETYP", imagetype, "Image type", "str")

        # update all headers with current data
        if self.update_headers_in_background:
            headerthread = threading.Thread(
                target=self.update_headers, name="updateheaders", args=[]
            )
            headerthread.start()
        else:
            self.update_headers()

        # flush detector
        if self.flush_array:
            self.flush()
        else:
            azcam.db.tools["controller"].stop_idle()

        # record current time and date in header
        self.record_current_times()

        return

    def integrate(self):
        """
        Integration.
        """

        return

    def readout(self):
        """
        Exposure readout.
        """

        return

    def end(self):
        """
        Completes an exposure by writing file and displaying image.
        """

        return

    def start_readout(self):
        """
        Start immediate readout of an exposing image.
        Returns immediately, not waiting for readout to finish.
        Really sets a flag which is read in expose().
        """

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["READ"]

        return

    def flush(self, Cycles=1):
        """
        Flush/clear detector.
        Cycles is the number of times to flush the detector.
        """

        azcam.log("Flushing")

        return azcam.db.tools["controller"].flush(int(Cycles))

    def sequence(self, number_exposures=-1, flush_array_flag=-1, delay=-1):
        """
        Take an exposure sequence.
        Uses pre-set exposure time, image type and image title.
        NumberExposures is the number of exposures to make.
        FlushArrayFlag defines detector flushing:
        -1 => current value defined by exposure.exposure_sequence_flush [default]
        0 => flush for each exposure
        1 => flush after first exposure only
        2 => no flush
        Delay => delay between exposures in seconds
        -1 => no change
        """

        AbortFlag = 0
        self.is_exposure_sequence = 1
        self.exposure_sequence_number = 1

        # number exposures in sequence
        number_exposures = int(number_exposures)
        if number_exposures == -1:
            number_exposures = self.exposure_sequence_total
        else:
            self.exposure_sequence_total = number_exposures

        # delay between exposures
        if int(delay) == -1:
            delay = self.exposure_sequence_delay
        delay = float(delay)

        # flushing
        flush_array_flag = int(flush_array_flag)
        currentflush = self.flush_array
        if flush_array_flag == -1:
            flush_array_flag = self.exposure_sequence_flush

        if flush_array_flag == 0 or flush_array_flag == 1:
            FlushArray = True
        else:
            FlushArray = False
        self.flush_array = FlushArray

        self.comp_sequence = (
            self.check_comparison_imagetype()
            and azcam.db.tools["instrument"].is_enabled
        )

        if self.comp_sequence:
            azcam.log("Starting comparison sequence")
            azcam.db.tools["instrument"].set_comps(self.image_type)
            if azcam.db.tools["instrument"].shutter_strobe:
                pass  # these instruments use shutter to turn on comps
            else:
                azcam.db.tools["instrument"].comps_on()
            azcam.db.tools[
                "instrument"
            ].comps_delay()  # delay for lamp warmup if needed

        for i in range(number_exposures):
            if i > 0:
                time.sleep(self.exposure_sequence_delay)

            if i > 0 and flush_array_flag == 1:
                self.flush_array = False

            self.expose(self.exposure_time, self.image_type, self.title)

            # check
            if azcam.db.abortflag:
                break

            # sequence may have been stopped
            if not self.is_exposure_sequence:
                break

            self.exposure_sequence_number += 1

        # turn off comps
        if self.comp_sequence:
            azcam.db.tools["instrument"].comps_off()
            self.comp_sequence = 0

        self.flush_array = currentflush
        self.is_exposure_sequence = 0
        self.exposure_sequence_number = 1

        azcam.db.abortflag = 0

        return

    def sequence1(self, number_exposures=1, flush_array_flag=-1, delay=-1):
        """
        Take an exposure sequence.
        Uses pre-set exposure time, image type and image title.
        number_exposures is the number of exposures to make.
        flush_array_flag defines detector flushing:
        -1 => current value defined by exposure.exposure_sequence_flush [default]
        0 => flush for each exposure
        1 => flush after first exposure only
        2 => no flush
        Delay => delay between exposures in seconds
        -1 => no change
        """

        arglist = [number_exposures, flush_array_flag, delay]
        thread = threading.Thread(target=self.sequence, name="sequence1", args=arglist)
        thread.start()

        return

    def pause(self):
        """
        Pause an exposure in progress (integration only).
        Really sets a flag which is read in expose().
        """

        self.paused_time_start = time.time()  # save paused clock time

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["PAUSE"]

        return

    def resume(self):
        """
        Resume a paused exposure.
        Really sets a flag which is read in expose().
        """

        self.paused_time = (
            time.time() - self.paused_time_start
        ) + self.paused_time  # total paused time in seconds

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["RESUME"]

        return

    def abort(self):
        """
        Abort an exposure in progress.
        Really sets a flag which is read in expose().
        """

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["ABORT"]

        return

    def set_shutter(self, state: int = 0, shutter_id: int = 0):
        """
        Open or close a shutter.

        :param state:
        :param shutter_id: Shutter ID flag

          * 0 => controller default shutter.
          * 1 => instrument default shutter.
        """

        if shutter_id == 0:
            azcam.db.tools["controller"].set_shutter(state)
        elif shutter_id == 1:
            azcam.db.tools["instrument"].set_shutter(state)

        return

    def parshift(self, number_rows: int):
        """
        Shift sensor by number_rows.

        :param number_rows: number of rows to shift (positive is toward readout, negative is away)
        """

        number_rows = int(float(number_rows))

        azcam.db.tools["controller"].parshift(number_rows)

        return

    # ***************************************************************************
    # pixel counter
    # ***************************************************************************

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        reply = azcam.db.tools["controller"].get_pixels_remaining()
        self.pixels_remaining = reply

        return reply

    # ***************************************************************************
    # exposure time
    # ***************************************************************************

    def get_exposuretime(self):
        """
        Return current exposure time in seconds.
        """

        if azcam.db.tools["controller"].is_reset:
            self.exposure_time = azcam.db.tools["controller"].get_exposuretime()

        return self.exposure_time

    def set_exposuretime(self, exposure_time: float):
        """
        Set current exposure time in seconds.
        """

        self.exposure_time = float(exposure_time)
        self.exposure_time_actual = self.exposure_time  # may be changed later

        azcam.db.tools["controller"].set_exposuretime(self.exposure_time)

        self.header.set_keyword(
            "EXPREQ", exposure_time, "Exposure time requested (seconds)", "float"
        )
        self.header.set_keyword(
            "EXPTIME", exposure_time, "Exposure time (seconds)", "float"
        )

        return

    def get_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        if azcam.db.tools["controller"].is_reset:
            self.exposure_time_remaining = azcam.db.tools[
                "controller"
            ].update_exposuretime_remaining()

        return self.exposure_time_remaining

    # **********************************************************************************************
    # header
    # **********************************************************************************************
    def record_current_times(self):
        """
        Record the current times and data info in the header.
        """

        # get current time and date
        self.obstime.update(0)

        # format should be YYYY-MM-DDThh:mm:ss.sss  ISO 8601
        self.header.set_keyword(
            "DATE-OBS", self.obstime.date[0], "UTC shutter opened", "str"
        )
        self.header.set_keyword(
            "DATE", self.obstime.date[0], "UTC date and time file writtten", "str"
        )  # OLD
        self.header.set_keyword(
            "TIME-OBS", self.obstime.ut[0], "UTC at start of exposure", "str"
        )
        self.header.set_keyword(
            "UTC-OBS", self.obstime.ut[0], "UTC at start of exposure", "str"
        )
        self.header.set_keyword(
            "UT", self.obstime.ut[0], "UTC at start of exposure", "str"
        )
        self.header.set_keyword(
            "TIMESYS", self.obstime.time_system[0], "Time system", "str"
        )
        self.header.set_keyword(
            "TIMEZONE", self.obstime.time_zone[0], "Local time zone", "str"
        )
        self.header.set_keyword(
            "LOCTIME",
            self.obstime.local_time[0],
            "Local time at start of exposure",
            "str",
        )

        return

    def update_headers(self):
        """
        Update all headers, reading current data.
        """

        # set flag that update is in progress
        self.updating_header = 1

        # all headers to be updated must be in azcam.db['headers']
        for objectname in azcam.db.headers:
            if (
                objectname == "controller"
                or objectname == "system"
                or objectname == "exposure"
                or objectname == "focalplane"
            ):
                continue
            try:
                azcam.db.tools[
                    objectname
                ].update_header()  # dont crash so all headers get updated
            except Exception as e:
                azcam.log(f"could not get {objectname} header: {e}")

        # update focalplane header which is not in db
        self.image.focalplane.update_header()

        # try to update system header last
        if "system" in azcam.db.headers:
            try:
                azcam.db.headers["system"].update_header()
            except Exception:
                pass

        # set flag that update is finished
        self.updating_header = 0

        return

    # ***************************************************************************
    # misc
    # ***************************************************************************

    def set_auto_title(self, flag=-1):
        """
        Set the AutoTitle flag so that image title matches lowercase imagetype.
        """

        if flag != -1:
            self.auto_title = flag
            self.set_image_title()

        if flag != 0:
            self.set_image_title()

        return

    def set_image_title(self, title=""):
        """
        Set the image title.
        Allows for AutoTitle.
        """

        if title == "":
            if self.title != "":
                title = self.title

        if self.auto_title:
            if (
                self.image_type.lower() == "object"
            ):  # don't change object title in AutoTitle mode
                pass
            else:
                # if title == "":
                #     title = self.image_type.lower()
                title = self.image_type.lower()

        # set OBJECT keyword to title or autotitle value
        self.set_keyword("OBJECT", title, "", "str")
        self.title = title
        self.image.title = title

        return

    def get_image_title(self):
        """
        Get the current image title.
        Allows for AutoTitle.
        """

        return self.title

    def set_image_type(self, imagetype="zero"):
        """
        Set image type for an exposure.
        imagetype is system defined, and typically includes:
        zero, object, dark, flat.
        """

        self.image_type = imagetype

        return

    def check_image_ready(self):
        """
        Returns True (1) if exposure image is ready and then toggles the flag off
        so that image is not detected multiple times.  Used only for clients which need to know
        when an image is ready.
        An image is ready when written to disk if SaveFile is true, or when valid if SaveFile is false.
        """

        if self.image.toggle:
            self.image.toggle = 0
            return 1
        else:
            return 0

    def get_image_type(self):
        """
        Get current image type for an exposure.
        imagetype is system defines, and typically includes:
        zero, object, dark, flat.
        """

        return self.image_type

    def get_image_types(self):
        """
        Return a list of valid imagetypes.
        Gets imagetypes and comparisons.
        """

        l1 = self.image_types

        try:
            l2 = azcam.db.tools["instrument"].get_all_comps()
        except Exception:
            return l1

        return l1 + l2 if len(l2) > 0 else l1

    def check_comparison_imagetype(self, imagetype=""):
        """
        Returns 1 if imagetype is a comparision, 0 if not.
        If imagetype is not specified, checks the current type.
        """

        if imagetype == "":
            imagetype = self.image_type  # if not specified use previous (global)

        # any undefined image types are comps
        if imagetype in self.image_types:
            return 0
        else:
            return 1

    def set_temp_files(self):
        """
        Update TempImageFile and TempDisplayFile file names based on CommandServer port.
        """

        if not os.path.isabs(self.temp_image_file):
            self.temp_image_file = os.path.join(azcam.db.datafolder, "TempImage")
            self.temp_image_file = os.path.normpath(self.temp_image_file)
            self.temp_image_file = "%s%d" % (
                self.temp_image_file,
                azcam.db.cmdserver.port,
            )  # make unique for multiple processes

        return

    def set_test_image(self, flag):
        """
        Sets TestImage flag is True, clears if False.
        """

        self.test_image = flag

        return

    def get_test_image(self):
        """
        Returns TestImage flag.
        """

        return self.test_image

    # **********************************************************************************************
    # detector parameters
    # **********************************************************************************************

    def set_detpars(self, sensor_data):
        """
        Set detector parameters from sensor data dictionary.
        """

        detpars = sensor_data

        if detpars.get("ref_pixel"):
            self.set_ref_pixel(detpars["ref_pixel"])
        if detpars.get("format"):
            self.set_format(*detpars["format"])
        if detpars.get("focalplane"):
            self.set_focalplane(*detpars["focalplane"])
        if detpars.get("roi"):
            self.set_roi(*detpars["roi"])
        if detpars.get("ext_position"):
            self.set_extension_position(detpars["ext_position"])
        if detpars.get("jpg_order"):
            self.set_jpg_order(detpars["jpg_order"])
        if detpars.get("det_number"):
            self.set_detnum(detpars["det_number"])
        if detpars.get("det_position"):
            self.set_detpos(detpars["det_position"])
        if detpars.get("det_gap"):
            self.set_detgap(detpars["det_gap"])
        if detpars.get("ext_name"):
            self.set_extname(detpars["ext_name"])
        if detpars.get("ext_number"):
            self.set_extnum(detpars["ext_number"])

        if detpars.get("ctype"):
            self.image.focalplane.wcs.ctype1 = detpars["ctype"][0]
            self.image.focalplane.wcs.ctype2 = detpars["ctype"][1]

        return

    def set_format(
        self,
        ns_total=-1,
        ns_predark=-1,
        ns_underscan=-1,
        ns_overscan=-1,
        np_total=-1,
        np_predark=-1,
        np_underscan=-1,
        np_overscan=-1,
        np_frametransfer=-1,
    ):
        """
        Set the detector format for subsequent exposures.
        Must call set_roi() after using this command and before starting exposure.
        ns_total is the number of visible columns.
        ns_predark is the number of physical dark underscan columns.
        ns_underscan is the desired number of desired dark underscan columns.
        ns_overscan is the number of dark overscan columns.
        np_total is the number of visible rows.
        np_predark is the number of physical dark underscan rows.
        np_underscan is the number of desired dark underscan rows.
        np_overscan is the number of desired dark overscan rows.
        np_frametransfer is the rows to frame transfer shift.
        """

        # update image.focalplane
        self.image.focalplane.set_format(
            ns_total,
            ns_predark,
            ns_underscan,
            ns_overscan,
            np_total,
            np_predark,
            np_underscan,
            np_overscan,
            np_frametransfer,
        )

        # update controller parameters
        controller = azcam.db.tools["controller"]
        controller.detpars.ns_total = self.image.focalplane.ns_total
        controller.detpars.ns_predark = self.image.focalplane.ns_predark
        controller.detpars.ns_underscan = self.image.focalplane.ns_underscan
        controller.detpars.ns_overscan = self.image.focalplane.ns_overscan
        controller.detpars.np_total = self.image.focalplane.np_total
        controller.detpars.np_predark = self.image.focalplane.np_predark
        controller.detpars.np_underscan = self.image.focalplane.np_underscan
        controller.detpars.np_overscan = self.image.focalplane.np_overscan
        controller.detpars.np_frametransfer = self.image.focalplane.np_frametransfer
        controller.detpars.coltotal = self.image.focalplane.ns_total
        controller.detpars.colusct = self.image.focalplane.ns_predark
        controller.detpars.coluscw = self.image.focalplane.ns_underscan
        controller.detpars.coluscm = 0
        controller.detpars.coloscw = self.image.focalplane.ns_overscan
        controller.detpars.coloscm = 0
        controller.detpars.rowtotal = self.image.focalplane.np_total
        controller.detpars.rowusct = self.image.focalplane.np_predark
        controller.detpars.rowuscw = self.image.focalplane.np_underscan
        controller.detpars.rowuscm = 0
        controller.detpars.rowoscw = self.image.focalplane.np_overscan
        controller.detpars.rowoscm = 0
        controller.detpars.framet = self.image.focalplane.np_frametransfer

        return

    def get_format(self):
        """
        Return the current detector format parameters.
        """

        return self.image.focalplane.get_format()

    def set_focalplane(
        self, numdet_x=-1, numdet_y=-1, numamps_x=-1, numamps_y=-1, amp_cfg=[0]
    ):
        """
        Sets focal plane configuration for subsequent exposures. Use after set_format().
        Must call set_roi() after using this command and before starting exposure.
        This command replaces SetConfiguration.
        Default focalplane values are set here.
        numdet_x defines number of detectors in Column direction.
        numdet_y defines number of detectors in Row direction.
        numamps_x defines number of amplifiers in Column direction.
        numamps_y defines number of amplifiers in Row direction.
        amp_cfg defines each amplifier's orientation (ex: [1,2,2,3]).
        0 - normal
        1 - flip x
        2 - flip y
        3 - flip x and y
        """

        # set type as could have some from socket string
        numdet_x = int(numdet_x)
        numdet_y = int(numdet_y)
        numamps_x = int(numamps_x)
        numamps_y = int(numamps_y)
        if type(amp_cfg) == str:
            amp_cfg = ast.literal_eval(amp_cfg)

        # update image.focalplane
        reply = self.image.focalplane.set_focalplane(
            numdet_x, numdet_y, numamps_x, numamps_y, amp_cfg
        )

        # update controller parameters
        controller = azcam.db.tools["controller"]
        controller.detpars.numdet_x = self.image.focalplane.numdet_x
        controller.detpars.numdet_y = self.image.focalplane.numdet_y
        controller.detpars.numamps_x = self.image.focalplane.numamps_x
        controller.detpars.numamps_y = self.image.focalplane.numamps_y
        controller.detpars.amp_cfg = self.image.focalplane.amp_cfg
        controller.detpars.num_detectors = self.image.focalplane.num_detectors
        controller.detpars.num_ser_amps_det = self.image.focalplane.num_ser_amps_det
        controller.detpars.num_par_amps_det = self.image.focalplane.num_par_amps_det
        controller.detpars.num_amps_det = self.image.focalplane.num_amps_det
        controller.detpars.numamps_image = self.image.focalplane.numamps_image
        controller.detpars.ampvispix_x = self.image.focalplane.ampvispix_x
        controller.detpars.ampvispix_y = self.image.focalplane.ampvispix_y

        self.image.set_scaling()

        return reply

    def get_focalplane(self):
        """
        Returns the current focal plane configuration.
        """

        return self.image.focalplane.get_focalplane()

    def set_data_order(self, dataOrder):
        """
        Sets data order
        """

        self.data_order = []
        if len(dataOrder) > 0:
            for item in dataOrder:
                self.data_order.append(int(item))

        return

    def set_roi(
        self,
        first_col=-1,
        last_col=-1,
        first_row=-1,
        last_row=-1,
        col_bin=-1,
        row_bin=-1,
        roi_num=0,
    ):
        """
        Sets the ROI values for subsequent exposures.
        Currently only one ROI (0) is supported.
        These values are for the entire focal plane, not just one detector.
        """

        # update image.focalplane
        first_col = int(first_col)
        last_col = int(last_col)
        first_row = int(first_row)
        last_row = int(last_row)
        col_bin = int(col_bin)
        row_bin = int(row_bin)
        roi_num = int(roi_num)
        self.image.focalplane.set_roi(
            first_col, last_col, first_row, last_row, col_bin, row_bin, roi_num
        )

        # update controller parameters
        controller = azcam.db.tools["controller"]
        controller.detpars.first_col = self.image.focalplane.first_col
        controller.detpars.last_col = self.image.focalplane.last_col
        controller.detpars.first_row = self.image.focalplane.first_row
        controller.detpars.last_row = self.image.focalplane.last_row
        controller.detpars.col_bin = self.image.focalplane.col_bin
        controller.detpars.row_bin = self.image.focalplane.row_bin

        controller.detpars.xunderscan = self.image.focalplane.xunderscan
        controller.detpars.xskip = self.image.focalplane.xskip
        controller.detpars.xpreskip = self.image.focalplane.xpreskip
        controller.detpars.xdata = self.image.focalplane.xdata
        controller.detpars.xpostskip = self.image.focalplane.xpostskip
        controller.detpars.xoverscan = self.image.focalplane.xoverscan
        controller.detpars.yunderscan = self.image.focalplane.yunderscan
        controller.detpars.yskip = self.image.focalplane.yskip
        controller.detpars.ypreskip = self.image.focalplane.ypreskip
        controller.detpars.ydata = self.image.focalplane.ydata
        controller.detpars.ypostskip = self.image.focalplane.ypostskip
        controller.detpars.yoverscan = self.image.focalplane.yoverscan

        controller.detpars.numcols_amp = self.image.focalplane.numcols_amp
        controller.detpars.numcols_overscan = self.image.focalplane.numcols_overscan
        controller.detpars.numviscols_amp = self.image.focalplane.numviscols_amp
        controller.detpars.numviscols_image = self.image.focalplane.numviscols_image
        controller.detpars.numrows_amp = self.image.focalplane.numrows_amp
        controller.detpars.numrows_overscan = self.image.focalplane.numrows_overscan
        controller.detpars.numvisrows_amp = self.image.focalplane.numvisrows_amp
        controller.detpars.numvisrows_image = self.image.focalplane.numvisrows_image
        controller.detpars.numpix_amp = self.image.focalplane.numpix_amp
        controller.detpars.numcols_det = self.image.focalplane.numcols_det
        controller.detpars.numrows_det = self.image.focalplane.numrows_det
        controller.detpars.numpix_det = self.image.focalplane.numpix_det
        controller.detpars.numpix_image = self.image.focalplane.numpix_image
        controller.detpars.numcols_image = self.image.focalplane.numcols_image
        controller.detpars.numrows_image = self.image.focalplane.numrows_image
        controller.detpars.numbytes_image = self.image.focalplane.numbytes_image
        controller.detpars.xflush = self.image.focalplane.xflush
        controller.detpars.yflush = self.image.focalplane.yflush

        # update controller
        controller.set_roi()

        # update image size
        self.size_x = self.image.focalplane.numcols_image
        self.size_y = self.image.focalplane.numrows_image

        self.image.size_x = self.image.focalplane.numcols_image
        self.image.size_y = self.image.focalplane.numrows_image

        # indicate that ROI has changed for next exposure
        self.new_roi = 1

        return

    def get_roi(self, roi_num=0):
        """
        Returns a list of the ROI parameters for the roi_num specified.
        Currently only one ROI (0) is supported.
        Returned list format is (first_col,last_col,first_row,last_row,col_bin,row_bin).
        """

        return self.image.focalplane.get_roi(roi_num)

    def roi_reset(self):
        """
        Resets detector ROI values to full frame, current binning.
        """

        self.image.focalplane.roi_reset()

        return

    def set_ref_pixel(self, XY):
        """
        Set the reference pixel.
        XY is [X,Y] in pixels.
        """

        self.image.focalplane.set_ref_pixel(XY)

        return

    def set_extension_position(self, XY):
        """
        Set the extension position of each amplifier.
        XY is [[X,Y]] in index numbers, starting at [1,1].
        """

        self.image.focalplane.set_extension_position(XY)

        return

    def set_detnum(self, det_number):
        """
        Set the detector numbers.
        """

        self.image.focalplane.det_number = det_number

        return

    def set_detgap(self, det_gap):
        """
        Set the detector gaps in pixels.
        """

        for i, gap in enumerate(det_gap):
            self.image.focalplane.gapx[i] = gap[0]
            self.image.focalplane.gapy[i] = gap[1]

        return

    def set_detpos(self, det_position):
        """
        Set the detector positions.
        """

        for i, pos in enumerate(det_position):
            self.image.focalplane.detpos_x[i] = pos[0]
            self.image.focalplane.detpos_y[i] = pos[1]

        return

    def set_jpg_order(self, Indices):
        """
        Set JPG image positions.
        """

        self.image.focalplane.set_jpg_order(Indices)

        return

    def set_tdi_delay(self, flag):
        """
        Set TDI line delay On or Off.
        flag==True sets delay to self.tdi_delay.
        flag==Flase sets line delay to normal value, self.par_delay.
        """

        return

    def set_extname(self, ext_name):
        """
        Set image extension names.
        """

        self.image.focalplane.set_extension_name(ext_name)

        return

    def set_extnum(self, ext_number):
        """
        Set image extension numbers.
        """

        self.image.focalplane.set_extension_extnum(ext_number)

        return

    def get_status(self):
        """
        Return a variety of system status data in one dictionary.
        """

        progress = 0

        filename = self.get_filename()
        et = self.get_exposuretime()
        if self.is_exposure_sequence:
            self.exposure_sequence_number
            seqcount = self.exposure_sequence_number
            seqtotal = self.exposure_sequence_total
        else:
            seqcount = 0
            seqtotal = 0

        ef = self.exposure_flag
        expstate = self.exposureflags_rev.get(ef, "")

        if azcam.db.tools["tempcon"].is_enabled:
            try:
                camtemp, dewtemp = azcam.db.tools["tempcon"].get_temperatures()[0:2]
                camtemp = f"{camtemp:.1f}"
                dewtemp = f"{dewtemp:.1f}"
            except Exception as e:
                # azcam.log(e)
                camtemp = -999.9  # error reading temperature
                dewtemp = -666.6
        else:
            camtemp = -999.9
            dewtemp = -666.6
            # camtemp = f"{camtemp:8.3f}"
            # dewtemp = f"{dewtemp:8.3f}"

        if ef == 1:
            expcolor = "green"
            et = self.get_exposuretime()
            if et == 0:
                progress = 0.0
                explabel = ""
            else:
                etr = self.get_exposuretime_remaining()
                explabel = f"{etr:.1f} sec remaining"
                progress = float(100.0 * (etr / et))
        elif ef == 7:
            expcolor = "red"
            progress = int(
                100.0
                * (self.get_pixels_remaining() / self.image.focalplane.numpix_image)
            )
            explabel = f"{progress}% readout"
        elif ef == 8:
            expcolor = "cyan"
            explabel = "exposure setup"
        else:
            expcolor = "transparent"
            progress = 0.0
            explabel = ""
            expstate = ""

        if self.message == "" and expstate != "":
            message = expstate
            if self.is_exposure_sequence:
                message = f"{message} - {self.exposure_sequence_number} of {self.exposure_sequence_total}"
        else:
            message = self.message

        # debug
        if 0:
            self.pgress += 5.0
            if self.pgress > 100:
                self.pgress = 0
            progress = self.pgress

        response = {
            "message": message,
            "exposurelabel": explabel,
            "exposurecolor": expcolor,
            "exposurestate": expstate,
            "progressbar": progress,
            "camtemp": camtemp,
            "dewtemp": dewtemp,
            "filename": filename,
            "seqcount": seqcount,
            "seqtotal": seqtotal,
            "timestamp": self._timestamp(0),
            "imagetitle": self.get_image_title(),
            "imagetype": self.get_image_type(),
            "imagetest": self.test_image,
            "exposuretime": self.get_exposuretime(),
            "colbin": self.image.focalplane.col_bin,
            "rowbin": self.image.focalplane.row_bin,
            "systemname": azcam.db.systemname,
            "mode": azcam.db.servermode,
        }

        return response

    def _timestamp(self, precision=0):
        """
        Return a timestamp string.
        precision is number of fractional seconds.
        """

        dateTimeObj = datetime.datetime.now()
        timestamp = str(dateTimeObj)

        if precision >= 6:
            pass
        elif precision == 0:
            timestamp = timestamp[:-7]
        else:
            tosecs = timestamp[:-7]
            frac = str(round(float(timestamp[-7:]), precision))
            timestamp = tosecs + frac

        return timestamp

    def read_header_file(self, filename):
        """
        Read header template file.
        """

        return azcam.db.headers["system"].read_file(filename)



"""
*azcam_console.utils* contains general purpose support commands used throughout azcam_console.
"""

import os
import shlex
import sys
from typing import List


# keyboard checking is optional
try:
    import msvcrt
except Exception:
    pass

import azcam
import azcam.exceptions


def curdir(folder: str = "") -> str:
    """
    Gets and sets the working folder.
    If folder is not specified then just return the current working folder.

    Args:
        folder: name of folder set.
    Returns:
        the current folder (after changing).
    """

    if folder is None:
        return

    if folder != "":
        folder = folder.lstrip('"').rstrip('"')
        try:
            os.chdir(folder)
        except FileNotFoundError:
            pass

    reply = os.getcwd()

    reply = reply.replace("\\", "/")

    azcam.db.wd = reply  # save result

    return reply


def fix_path(path: str = "", no_drive_letter: bool = True) -> str:
    """
    Makes a nice absolute path, leaving only forward slashes.

    Args:
        path: name of path to cleanup.
        no_drive_letter: Removes leading drive letter.
    Returns:
        cleaned path name.
    """

    if os.name != "posix":
        norm = os.path.abspath(os.path.normpath(path))
        pth = norm.replace("\\", "/")  # go to forward slashes only
        if no_drive_letter and len(pth) > 2 and pth[1] == ":":
            pth = pth[2:]
    else:
        pth = os.path.expanduser(path)
        pth = os.path.normpath(pth)

    return pth


def add_searchfolder(search_folder: str = "", include_subfolders: bool = True) -> None:
    """
    Appends search_folder (and by default all its subfolders) to the current python search path.
    Default is current folder and its subfolders.
    Subfolders beginning with "_" are not included.

    Args:
        search_folder: Name of folder to add to sys.path
        include_subfolders: True to include all subfolders in sys.path
    """

    if search_folder == "":
        search_folder = curdir()

    search_folder = fix_path(search_folder)

    # append all subfolders of search_folder to current search path
    if search_folder not in sys.path:
        sys.path.append(search_folder)

    if include_subfolders:
        for root, dirs, _ in os.walk(search_folder):
            if dirs:
                for s in dirs:
                    if s.startswith("_"):
                        continue
                    sub = os.path.join(root, s)
                    sub = fix_path(sub)
                    if sub not in sys.path:
                        sys.path.append(sub)

    return


def make_image_filename(imagefile: str) -> str:
    """
    Returns the absolute file imagefile, with forward slashes.
    Appends ".fits" if no extension is included.

    Args:
        imagefile: image filename to be expanded
    Returns:
        expanded image filename.
    """

    if imagefile.endswith(".fits"):
        pass
    elif imagefile.endswith(".fit"):
        pass
    elif not imagefile.endswith(".bin"):
        imagefile += ".fits"

    return fix_path(imagefile)


def parse(string: str, set_type: bool = False) -> List[str]:
    """
    Parse a string into tokens using the standard azcam rules.
    If setType is true, try and set data data type for each token.

    Args:
        string: String to be parsed into tokens
        set_type: True to try and set the type of each token ("1" to 1)
    Returns:
        list of parsed tokens
    """

    # allow for quotes
    lex = shlex.shlex(string)
    lex.quotes = "\"'"
    lex.whitespace_split = True
    lex.commenters = "#"
    toks = list(lex)

    # remove bounding quotes unless quoting a number (leave as string)
    tokens = []
    for tok in toks:
        if tok.startswith('"') and tok.endswith('"'):
            tok1 = tok[1:-1]
            # t, value = get_datatype(tok1)
            # if t not in ["int", "float"]:
            #     tok = tok1

        elif tok.startswith("'") and tok.endswith("'"):
            tok1 = tok[1:-1]
            # t, value = get_datatype(tok1)
            # if t not in ["int", "float"]:
            #     tok = tok1

        tokens.append(tok)

    if set_type:
        for i, tok in enumerate(tokens):
            t, value = get_datatype(tok)
            tokens[i] = value

    return tokens


def get_datatype(value: any) -> list:
    """
    Determine the data type for an object and set the type if possible. A string such as "1.23"
    will result in a type "float" and "2" will result in type "int".

    Args:
        value: object to be typed
    Returns:
        list [type, value] of data type as a code and object with that type
    """

    if isinstance(value, str):
        # string integer
        if value.isdigit():
            attributetype = "int"
            value = int(value)
            return [attributetype, value]
        else:
            try:
                value = float(value)
                attributetype = "float"
                return [attributetype, value]
            except ValueError:
                pass

        attributetype = "str"

    elif isinstance(value, int):
        attributetype = "int"
        value = int(value)

    elif isinstance(value, float):
        attributetype = "float"
        value = float(value)

    # more work here
    else:
        attributetype = "str"

    return [attributetype, value]


def prompt(prompt_message: str = "Enter a string", default: any = "") -> any:
    """
    Prints a message and waits for user input.

    Args:
        prompt_message: string to be printed
        default:  string to be returned if no value is entered
    Returns:
        string entered or default value
    """

    default = str(default)
    try:
        if default != "":
            in1 = input(prompt_message + " [" + default + "]: ")
        else:
            in1 = input(prompt_message + ": ")
    except KeyboardInterrupt:
        return ""

    if in1 == "":
        return default
    else:
        return in1


def check_keyboard(wait: bool = False) -> str:
    """
    Checks keyboard for a key press.
    For Windows OS only.

    Args:
        wait: True to wait until a key is pressed
    Returns:
        key which was pressed or empty string.
    """

    # TODO: map sequences like 'F1'

    if os.name != "nt":
        # raise azcam.exceptions.AzcamError("check_keyboard not supported on this OS")
        return ""

    loop = 1
    key = ""

    while loop:
        if msvcrt.kbhit():
            key = msvcrt.getch()
            try:
                key = key.decode()

                # since the key is byte type, maybe escape sequence so check for more
                # if msvcrt.kbhit():
                #    key1 = msvcrt.getch()
                #    # key = key + key1.decode()
            except UnicodeDecodeError:
                pass
            break
        if not wait:
            loop = 0

    return key


def show_menu(configs: dict) -> str:
    """
    Interative: Show a menu and wait for selection.
    "blank" may be used to display an empty line.
    print() is allowed here as this is for interactive use only.

    Args:
        configs: Dictionary of strings which are menu items
    Returns:
        string associated with item selected or empty string.
    """

    if len(configs) == 1:
        choice = configs[list(configs.keys())[0]]
        return choice

    CONFIRMED = 0
    choice = ""
    while not CONFIRMED:
        print("Select configuration number from list below:\n")
        i = 0
        for c in configs:
            if c == "blank":
                print("")
            else:
                i += 1
                print("%1d.....%s" % (i, c))
        print("")
        print("Enter configuration number: ", end="")
        choiceindex = input()
        if choiceindex == "q":
            azcam.exceptions.warning("Quit detected")
            return
        try:
            choiceindex = int(choiceindex)
        except ValueError:
            print("Bad keyboard input...try again\n")
            continue

        choiceindex = int(choiceindex)
        choiceindex = choiceindex - 1  # zero based

        # remove blanks
        for x in configs:
            if x == "blank":
                configs.remove("blank")

        if choiceindex < 0 or choiceindex > len(configs) - 1:
            print("invalid selection - %d\n" % (choiceindex + 1))
            continue

        # get choice
        configlist = list(configs.keys())  # is order OK?
        choice = configs[configlist[choiceindex]]

        CONFIRMED = 1

    print("")

    return choice


def get_datafolder(datafolder: str | None = None):
    """
    Return the datafolder for this system.
    If not specified, root is /data on Windows or ~/data on Linux.
    """

    if datafolder is None:
        droot = os.environ.get("AZCAM_DATAROOT")
        if droot is None:
            if os.name == "posix":
                droot = os.environ.get("HOME")
            else:
                droot = "/"
            datafolder = os.path.join(
                os.path.realpath(droot), "data", azcam.db.systemname
            )
        else:
            datafolder = os.path.join(os.path.realpath(droot), azcam.db.systemname)
    else:
        datafolder = os.path.realpath(datafolder)

    datafolder = os.path.normpath(datafolder)

    return datafolder


def save_imagepars(imagepars: dict) -> None:
    """
    Save current image parameters.
    imagepars is a dictionary.

    Args:
        imagepars: dict of azcam.db.imageparnames names to save
    """

    for par in azcam.db.imageparnames:
        imagepars[par] = azcam.db.parameters.get_par(par)

    return


def restore_imagepars(imagepars: dict) -> None:
    """
    Restore image parameters from dictionary.

    Args:
        imagepars: dictionary set with save_imagepars().
    """

    for par in azcam.db.imageparnames:
        value = quoter(imagepars[par])
        imagepars[par] = value
        azcam.db.parameters.set_par(par, value)

    return


def dequote(input: str) -> str:
    """
    Remove matching single or double quote at ends of input.

    Args:
        input: string to be dequoted.
    """

    if type(input) != str:
        return input

    if input.startswith("'") and input.endswith("'"):
        dequote = input[1:-1]
    elif input.startswith('"') and input.endswith('"'):
        dequote = input[1:-1]
    else:
        dequote = input

    return dequote


def quoter(input: str) -> str:
    """
    Put quotes around a string if it contains special characters.
    Currently just spaces.

    Args:
        input: string to be optionally quoted.
    """

    if type(input) != str:
        return input

    if " " not in input:
        return input

    if input.startswith("'") and input.endswith("'"):
        output = input
    elif input.startswith('"') and input.endswith('"'):
        output = input
    else:
        output = f'"{input}"'

    return output



"""
Contains the main azcam database class.
"""

from typing import Any, Union, List, Dict

import azcam
from azcam.logger import AzCamLogger
from azcam.parameters import Parameters


class AzcamDatabase(object):
    """
    The azcam database class.
    """

    version: str = azcam.__version__
    """current azcam version"""

    wd: Union[str, None] = None
    """the current working directory"""

    verbosity: int = 1
    """verbosity level for messages"""

    mode: str = ""
    """operating mode (server or console)"""

    abortflag: int = 0
    """abort flag, 1 (true) if an abort has occurred"""

    datafolder: str = ""
    """system datafolder"""

    systemname: str = ""
    """system name"""

    systemfolder: str = ""
    """system root folder"""

    tools: dict = {}
    """dict of tools"""

    cli: dict = {}
    """dict of CLI commands"""

    scripts: dict = {}
    """dict of scripts"""

    headers: dict = {}
    """header objects"""

    headerorder: list = []
    """header order in image header"""

    logger: AzCamLogger = AzCamLogger()
    """logger object"""

    parameters: Parameters
    """parameters object"""

    # *************************************************************************
    # parameter table
    # *************************************************************************
    par_table: dict = {
        # exposure
        "autotitle": "exposure.auto_title",
        "imagetype": "exposure.image_type",
        "exposureflag": "exposure.exposure_flag",
        "exposuresequencedelay": "exposure.exposure_sequence_delay",
        "exposuresequencetotal": "exposure.exposure_sequence_total",
        "exposuresequencenumber": "exposure.exposure_sequence_number",
        "exposuresequenceflush": "exposure.exposure_sequence_flush",
        "exposureupdatingheader": "exposure.updating_header",
        "isexposuresequence": "exposure.is_exposure_sequence",
        "displayimage": "exposure.display_image",
        "sendimage": "exposure.send_image",
        "savefile": "exposure.save_file",
        "flusharray": "exposure.flush_array",
        "tdidelay": "exposure.tdi_delay",
        "tdimode": "exposure.tdi_mode",
        "pardelay": "exposure.par_delay",
        "exposureguidemode": "exposure.guide_mode",
        "exposureguidestatus": "exposure.guide_status",
        "lastfilename": "exposure.last_filename",
        "imagefiletype": "exposure.filetype",
        "imageheaderfile": "exposure.imageheaderfile",
        "imagetest": "exposure.test_image",
        "imagesequencenumber": "exposure.sequence_number",
        "imageautoincrementsequencenumber": "exposure.auto_increment_sequence_number",
        "imageincludesequencenumber": "exposure.include_sequence_number",
        "imageautoname": "exposure.autoname",
        "imageoverwrite": "exposure.overwrite",
        "imageroot": "exposure.root",
        "imagefolder": "exposure.folder",
        # image
        "imagesizex": "exposure.image.focalplane.numcols_image",
        "imagesizey": "exposure.image.focalplane.numrows_image",
        "numpiximage": "exposure.image.focalplane.numpix_image",
        "colbin": "exposure.image.focalplane.col_bin",
        "rowbin": "exposure.image.focalplane.row_bin",
        "firstcol": "exposure.image.focalplane.first_col",
        "firstrow": "exposure.image.focalplane.first_row",
        "lastcol": "exposure.image.focalplane.last_col",
        "lastrow": "exposure.image.focalplane.last_row",
        # instrument
        "instrumentenabled": "instrument.is_enabled",
        "instrumentfocus": "instrument.focus_position",
        # telescope
        "telescopeenabled": "telescope.is_enabled",
        "telescopefocus": "telescope.focus_position",
        # tempcon
        "controltemperature": "tempcon.control_temperature",
        "camtemp": "tempcon.temperatures[0]",
        "dewtemp": "tempcon.temperatures[1]",
        # controller
        "utilityboardinstalled": "controller.utility_board_installed",
        "pciboardinstalled": "controller.pci_board_installed",
        "timingboardinstalled": "controller.timing_board_installed",
        "videogain": "controller.video_gain",
        "videospeed": "controller.video_speed",
        "usereadlock": "controller.use_read_lock",
        "pcifile": "controller.pci_file",
        "timingfile": "controller.timing_file",
        "utilityfile": "controller.utility_file",
        "timingboard": "controller.timing_board",
        "videoboards": "controller.video_boards",
        "clockboards": "controller.clock_boards",
        # database
        "systemname": "db.systemname",
        "abortflag": "db.abortflag",
        "verbosity": "db.verbosity",
        "hostname": "db.hostname",
        "version": "db.version",
    }
    """dict of general parameters"""

    imageparnames: List[str] = [
        "imageroot",
        "imageincludesequencenumber",
        "imageautoname",
        "imageautoincrementsequencenumber",
        "imagetest",
        "imagetype",
        "imagetitle",
        "imageoverwrite",
        "imagefolder",
    ]
    """image parameters"""

    exposureflags: Dict[str, int] = {
        "NONE": 0,
        "EXPOSING": 1,
        "ABORT": 2,
        "PAUSE": 3,
        "RESUME": 4,
        "READ": 5,
        "PAUSED": 6,
        "READOUT": 7,
        "SETUP": 8,
        "WRITING": 9,
        "GUIDEERROR": 10,
        "ERROR": 11,
    }
    """exposure flags"""

    # *************************************************************************
    # db methods
    # *************************************************************************

    def __init__(self) -> None:
        pass

    def get(self, name: str) -> Any:
        """
        Returns a database attribute by name.
        Args:
          name: name of attribute to return
        Returns:
          value or None if *name* is not defined
        """

        try:
            obj = getattr(self, name)
        except AttributeError:
            obj = None

        return obj

    def set(self, name: str, value: Any) -> None:
        """
        Sets a database attribute value.
        Args:
          name: name of attribute to set
          value: value of attribute to be set
        """

        # if not hasattr(self, name):
        #    return

        setattr(self, name, value)

        return



"""
server script to test Mag controller - server-side
"""

import sys

import azcam


def test_mag_controller(cycles=100):

    # loop through boards for communications
    azcam.log("Testing communications...")
    for loop in range(cycles):
        azcam.db.tools["controller"].test_datalink(2, 222, 1)
    azcam.log("Communications to controller OK")

    # reset
    azcam.log("Testing controller reset...")
    for loop in range(cycles):
        azcam.db.tools["controller"].reset()
        azcam.log(f"Controller reset {loop}/{cycles}")

    # reset
    azcam.log("Testing exposures...")
    for loop in range(cycles):
        azcam.db.tools["exposure"].expose()
        azcam.log(f"Exposure {loop}/{cycles}")

    return


if __name__ == "__main__":
    args = sys.argv[1:]
    test_mag_controller(*args)



"""
Contains the ExposureASCOM class.
"""

import os
import time
import numpy

import azcam
from azcam.tools.exposure import Exposure


class ExposureASCOM(Exposure):
    """
    Defines the exposure class for cameras using ASCOM.
    """

    def __init__(self, tool_id="exposure", description=None):
        super().__init__(tool_id, description)

        self.exp_start = 0
        self.readout_delay = 0

    def integrate(self):
        """
        Integration.
        """

        self.exposure_flag = self.exposureflags["EXPOSING"]
        imagetype = self.image_type.lower()

        try:
            shutterstate = self.shutter_dict[self.image_type.lower()]
        except KeyError:
            shutterstate = 1  # other types are comps, so open shutter

        # close instrument shutter for darks and zeros
        if azcam.db.tools["instrument"].is_enabled:
            if shutterstate:
                azcam.db.tools["instrument"].comps_on()
            else:
                azcam.db.tools["instrument"].comps_off()

        # start exposure
        if imagetype != "zero":
            azcam.log("Integration started")

        self.exp_start = time.time()
        self.dark_time_start = self.exp_start
        azcam.db.tools["controller"].start_exposure()

        # wait for integration
        time.sleep(self.exposure_time)

        # exposure finished
        if imagetype == "zero":
            self.exposure_time = self.exposure_time_saved
        self.exposure_time_remaining = 0
        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.log("Integration aborted")
        else:
            self.exposure_flag = self.exposureflags["READ"]

        if azcam.db.tools["instrument"].is_enabled:
            azcam.db.tools["instrument"].comps_off()

        if self.image_type.lower() != "zero":
            azcam.log("integration finished", level=2)

        return

    def readout(self):
        """
        Exposure readout.
        """

        self.exposure_flag = self.exposureflags["READOUT"]

        # wait for readout
        azcam.log("Readout started")

        try:
            azcam.db.tools["controller"].is_imageready(1)
        except azcam.exceptions.AzcamError:
            self.exposure_flag = self.exposureflags["ABORT"]
            return

        # allow readout to complete
        self.image.valid = 1

        imagetype = self.image_type.lower()
        if imagetype == "ramp":
            azcam.db.tools["controller"].set_shutter_state(0)

        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.log("Readout aborted")
            return
        else:
            azcam.log("Readout finished", level=2)
            self.exposure_flag = self.exposureflags["NONE"]
            return

    def end(self):
        """
        Completes an exposure by writing file and displaying image.
        """

        # this can be slow for big image
        # ASCOM fills array Y then X do transpose
        self.image.transposed_image = 1
        # size = azcam.db.tools["controller"].camera.NumX * azcam.db.tools["controller"].camera.NumY  # can be wrong
        size = (
            azcam.db.tools["controller"].detpars.numcols_image
            * azcam.db.tools["controller"].detpars.numrows_image
        )
        self.image.data[0] = (
            numpy.array(azcam.db.tools["controller"].camera.ImageArray)
            .reshape(size)
            .astype("uint16")
        )

        self.exposure_flag = self.exposureflags["WRITING"]

        if self.send_image:
            local_file = self.temp_image_file + "." + self.get_extname(self.filetype)
            try:
                os.remove(local_file)
            except FileNotFoundError:
                pass
        else:
            local_file = self.get_filename()

        self.last_filename = local_file

        # update controller header with keywords which might have changed
        et = float(int(self.exposure_time_actual * 1000.0) / 1000.0)
        dt = float(int(self.dark_time * 1000.0) / 1000.0)
        azcam.db.headers["exposure"].set_keyword(
            "EXPTIME", et, "Exposure time (seconds)", "float"
        )
        azcam.db.headers["exposure"].set_keyword(
            "DARKTIME", dt, "Dark time (seconds)", "float"
        )

        # write file(s) to disk
        if self.save_file:
            azcam.log("Writing %s" % local_file)

            # write the file to disk
            self.image.overwrite = self.overwrite
            self.image.test_image = self.test_image

            self.image.write_file(local_file, self.filetype)

            azcam.log("Writing finished", level=2)

            # set flag that image now written to disk
            self.image.is_written = 1

            if self.send_image:
                azcam.log("Sending image")
                self.sendimage.send_image()

        # image data and file are now ready
        self.image.toggle = 1

        # display image
        if self.display_image:
            try:
                azcam.log("Displaying image")
                azcam.db.tools["display"].display(self.image)
            except Exception:
                pass

        # increment file sequence number if image was written
        if self.save_file:
            self.increment_filenumber()

        self.exposure_flag = self.exposureflags["NONE"]

        return

    def get_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        return azcam.db.tools["controller"].update_exposuretime_remaining()



"""
Contains custom exceptions and warnings used throughout azcam.
"""

import azcam

import warnings

warnings.filterwarnings("ignore")


def warning(message: str) -> None:
    """
    Print and log a warning message.
    """

    # warnings.warn(message)
    # print(f"Warning: {message}")

    try:
        azcam.logger.warning(f"Warning: {message}")
    except Exception:
        print(f"Warning: {message}")

    return


class AzcamError(Exception):
    """
    Base custom error class for azcam.
    """

    def __init__(self, message: str, error_code: int = 0):
        """
        Custom error exception for azcam.

        Usage:  raise azcam.exceptions.AzcamError(message)

        Args:
          message: string message to display when error is raised
          error_code: flag for code, from list below
          - 0 - no error
          - 1 - controller reset error, check power and fibers
          - 2 - could not open connection to server
          - 3 - receive image data abort
          - 4 - remote call not allowed
        """

        super().__init__(message)

        self.error_code = 0

        # Now for your custom code...
        if error_code is not None:
            self.error_code = error_code
            # Original error was self.errors.message

        if azcam.logger is not None:
            azcam.logger.error(message)
        else:
            print(f"AzcamError: {message}")



"""
test ARC controller - server-side
"""

import sys

import azcam


def test_arc_controller(cycles: int = 10):
    """
    Test ARC controller.
    """

    cycles = int(cycles)

    # loop through boards for communications
    azcam.log("Testing controller board level communications...")

    if azcam.db.tools["controller"].pci_board_installed:
        azcam.log("Testing PCI board...")
        for loop in range(cycles):
            azcam.db.tools["controller"].test_datalink(1, 111, 1)
        azcam.log("PCI board OK")

    if azcam.db.tools["controller"].timing_board_installed:
        azcam.log("Testing Timing board...")
        for loop in range(cycles):
            azcam.db.tools["controller"].test_datalink(2, 222, 1)
        azcam.log("Timing board OK")

    if azcam.db.tools["controller"].utility_board_installed:
        azcam.log("Testing Utility board...")
        for loop in range(cycles):
            azcam.db.tools["controller"].test_datalink(3, 333, 1)
        azcam.log("Utility board OK")

    # reset
    azcam.log("Testing controller reset...")
    for loop in range(cycles):
        azcam.db.tools["controller"].reset()
    azcam.log("Controller reset OK")

    # temperature
    if azcam.db.tools["controller"].utility_board_installed:
        azcam.log("Testing controller temperature reading...")
        for loop in range(cycles):
            azcam.db.tools["tempcon"].get_temperatures()
        azcam.log("Controller temerature reading OK")

    return


if __name__ == "__main__":
    args = sys.argv[1:]
    test_arc_controller(int(*args))



"""
UDP interface class
Allows sending UDP register request to the Azcam Monitor application
Allows sending and receiving broadcast ID request
"""

import socket
import time

import azcam

"""
Example:

.get_ip('guider_z1')
=> '10.0.1.100'

.get_ids()
=>[(b'0 4501 guider_z1 2425 10.0.100 0\r\n', ('10.0.1.100', 2400))]
"""


class UDPinterface(object):
    def __init__(self):

        self.resp = []
        self.wait = 1  # seconds

    def get_ip(self, hostName):
        """
        Sends UDP Get ID request (port 2400) and looks for a hostName,
        then returns IP address if found.
        """

        azcam.log("Resolving " + hostName + " IP Address")

        # create a ne wsocket for receiving IDs
        udp_socket_data = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        udp_socket_data.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_socket_data.setblocking(0)
        udp_socket_data.bind(("", 2401))

        # ID request
        cmd = "0\r\n"

        # create a new socket for sending register command
        udp_socket_ctrl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        udp_socket_ctrl.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_socket_ctrl.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # send ID request
        udp_socket_ctrl.sendto(bytes(cmd, "utf-8"), ("255.255.255.255", 2400))

        # close socket
        udp_socket_ctrl.close()

        # wait self.wait time for the responses
        start = time.time()
        loop_ok = True

        while loop_ok:

            stop = time.time()
            if stop - start > self.wait:
                loop_ok = False

            try:
                recv = udp_socket_data.recvfrom(1024)

                # store the whole response
                self.resp.append(recv)
            except Exception:
                pass

        # close socket
        udp_socket_data.close()

        # check IDs
        cnt = len(self.resp)
        ip_address = "0.0.0.0"

        if cnt > 0:
            for indx in self.resp:
                recv = str(indx).split(" ")
                try:
                    if recv[2] == hostName:
                        ip_address = recv[4]
                except Exception:
                    pass
        else:
            azcam.log("ERROR No IDs available")

        return ip_address

    def get_ids(self):
        """
        Sends UDP Get ID request (port 2400).
        """

        self.resp = []

        # create a new socket for receiving IDs
        udp_socket_data = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        udp_socket_data.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_socket_data.setblocking(0)
        udp_socket_data.bind(("", 2401))

        # ID request
        cmd = "0\r\n"

        # create a new socket for sending register command
        udp_socket_ctrl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        udp_socket_ctrl.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_socket_ctrl.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # send ID request
        udp_socket_ctrl.sendto(bytes(cmd, "utf-8"), ("255.255.255.255", 2400))

        # close socket
        udp_socket_ctrl.close()

        # wait self.wait time for the responses
        start = time.time()
        loop_ok = True

        while loop_ok:

            stop = time.time()
            if stop - start > self.wait:
                loop_ok = False

            try:
                recv = udp_socket_data.recvfrom(1024)
                # store the whole response
                self.resp.append((recv[0].decode(), recv[1]))
            except Exception:
                pass

        # close socket
        udp_socket_data.close()

        return self.resp



"""
CLI shortcuts for azcamserver.
"""

import azcam


def sf_server():
    """Shortcut to Set image folder"""

    try:
        folder = azcam.utils.curdir()
        azcam.db.tools["exposure"].folder = folder
    except Exception:
        pass

    return


def gf_server():
    """
    Shortcut to Go to image folder.
    Also issues sav() command to save folder location.
    """

    folder = azcam.db.tools["exposure"].folder
    azcam.utils.curdir(folder)
    azcam.db.wd = folder
    sav_server()

    return


def sav_server():
    """Shortcut for parfile_write() saving current folder in database."""
    azcam.db.parameters.set_par("wd", azcam.utils.curdir())
    azcam.db.parameters.update_par_dict()
    azcam.db.parameters.write_parfile()

    return None


def pp():
    """Shortcut to toggle cmdserver printing."""

    old = azcam.db.cmdserver.logcommands
    new = not old
    azcam.db.cmdserver.logcommands = new

    print(f"logcommands is now {'ON' if new else 'OFF'}")

    return


azcam.db.cli.update({"sav": sav_server, "pp": pp, "sf": sf_server, "gf": gf_server})



"""
Contains the ExposureArchon class.
"""

import os
import time

import numpy

import azcam
import azcam.exceptions
from azcam.tools.exposure import Exposure
from astropy.io import fits as pyfits


class ExposureArchon(Exposure):
    """
    The exposure class for Archon controllers.
    The methods here override the default Exposure class and are for the STA Archon controller
    used in 'direct mode', communicating directly with the controller hardware.
    """

    def __init__(self, tool_id="exposure", description=None):
        super().__init__(tool_id, description)

        self.receive_data = ReceiveDataArchon(self)
        self.fileconverter = ArchonFileConverter()

        # add extra extensions for additional non-image data
        self.add_extensions = 0

        # shutter delay in msec
        self.shutter_delay = 10

    def abort(self):
        """
        Abort an exposure in progress.
        Sends RESETTIMING to controller and sets the ExposureFlag which is read
        in during the exposure.
        """

        azcam.db.tools["controller"].resettiming()

        azcam.db.abortflag = 1

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["ABORT"]

        return

    def start_readout(self):
        """
        Immediate exposure readout.
        """

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["READ"]

        azcam.db.tools["controller"].archon_command("FASTLOADPARAM StopExposure 1")
        time.sleep(0.1)
        azcam.db.tools["controller"].archon_command("FASTLOADPARAM StopExposure 0")

        return

    def flush(self, cycles=1):
        """
        Flush/clear detector.
        cycles is the number of times to flush the detector.
        """

        # not supported yet
        return

    def integrate(self):
        """
        Integration.
        """

        # make the exposure itself here
        self.dark_time_start = time.time()
        # azcam.log("Integration started")

        azcam.db.tools["controller"].start_exposure(1)

        # dark time includes readout
        self.dark_time = time.time() - self.dark_time_start

        # turn off comp lamps
        if self.comp_exposure:
            if not azcam.db.tools["instrument"].shutter_strobe:
                azcam.db.tools["instrument"].comps_off()
            azcam.db.tools["instrument"].set_comps("shutter")

        # set times
        self.exposure_time_remaining = 0
        if self.image_type == "zero":
            self.exposure_time = self.exposure_time_saved

        self.exposure_flag == self.exposureflags["READ"]

        # azcam.log("Integration finished", level=2)

        return

    def end(self):
        """
        Completes an exposure by writing file and displaying image.
        """

        self.exposure_flag = self.exposureflags["WRITING"]

        if self.send_image:
            LocalFile = self.temp_image_file + "." + self.get_extname(self.filetype)
            try:
                os.remove(LocalFile)
            except FileNotFoundError:
                pass
        else:
            LocalFile = self.get_filename()

        self.last_filename = LocalFile

        # get the image data and put into controller.imagedata
        self.receive_data.receive_archon_image_data()

        self.pixels_remaining = 0

        # create buffer for entire image, including all overscans
        self.image.data = numpy.empty(
            shape=(
                self.image.focalplane.numamps_image,
                self.image.focalplane.numcols_amp * self.image.focalplane.numrows_amp,
            ),
            dtype="uint16",
        )

        self.fileconverter.copy_to_buffer(
            azcam.db.tools["controller"].imagedata, self.image.data
        )
        self.image.is_valid = 1

        # write MEF file
        self.image.overwrite = self.overwrite
        self.image.test_image = self.test_image

        azcam.db.tools["controller"].set_keyword(
            "INTMS", self.fileconverter.intms, "Open shutter exposure time (ms)", "int"
        )
        azcam.db.tools["controller"].set_keyword(
            "NOINTMS",
            self.fileconverter.nointms,
            "Closed shutter exposure time (ms)",
            "int",
        )

        # update controller header with keywords which might have changed
        et = float(int(self.exposure_time_actual * 1000.0) / 1000.0)
        self.dark_time = et  # does not yet include pause/resume
        dt = float(int(self.dark_time * 1000.0) / 1000.0)
        azcam.db.headers["exposure"].set_keyword(
            "EXPTIME", et, "Exposure time (seconds)", "float"
        )
        azcam.db.headers["exposure"].set_keyword(
            "DARKTIME", dt, "Dark time (seconds)", "float"
        )

        self.image.write_file(LocalFile, self.filetype)

        # add info data in extra extensions
        if self.add_extensions:
            azcam.db.tools["controller"].get_status()  # get current controller data

            # create data arrays
            keywords = []
            values = []
            datatypes = []
            units = []
            comments = []
            for key in azcam.db.tools["controller"].dict_status:
                keywords.append(key)
                values.append(azcam.db.tools["controller"].dict_status[key])
                datatypes.append("datatype")
                units.append("unit")
                comments.append("comment")

            # make table columns
            # keywords1 = numpy.array(keywords)
            # values1 = numpy.array(values)
            c1 = pyfits.Column(name="Keyword", format="20A", array=keywords)
            c2 = pyfits.Column(name="Value", format="20A", array=values)
            # c3 = pyfits.Column(name="DataType", format="20A", array=datatypes)
            # c4 = pyfits.Column(name="Units", format="20A", array=units)
            # c5 = pyfits.Column(name="Comment", format="80A", array=comments)
            # coldefs = pyfits.ColDefs([c1, c2, c3, c4, c5])
            coldefs = pyfits.ColDefs([c1, c2])

            hdu_conpars = pyfits.BinTableHDU.from_columns(coldefs, name="CONPARS")
            with pyfits.open(LocalFile, mode="update") as hdulist:
                hdulist.append(hdu_conpars)

        azcam.log(f"Writing finished: {LocalFile}", level=2)

        # set flag that image now written to disk
        self.image.is_written = 1

        # image data and file are now ready
        self.image.toggle = 1

        # display image
        if self.display_image:
            azcam.log("Displaying image")
            azcam.db.tools["display"].display(LocalFile)

        if self.send_image:
            azcam.log("Sending image")
            self.sendimage.send_image(LocalFile, self.get_filename())

        # increment file sequence number if image was written
        if self.save_file:
            self.increment_filenumber()

        self.exposure_flag = self.exposureflags["NONE"]

        return

    def set_exposuretime(self, exposure_time):
        """
        Set current exposure time in seconds.
        """

        self.exposure_time = float(exposure_time)
        self.exposure_time_actual = self.exposure_time  # may be changed later

        if self.image_type == "zero":
            azcam.db.tools["controller"].set_exposuretime(0)
            azcam.db.tools["controller"].set_no_int_ms(0)
            return

        # set timer and for header keyword
        azcam.db.tools["controller"].set_exposuretime(int(self.exposure_time * 1000))

        # get shutter state
        try:
            shutterstate = self.shutter_dict[self.image_type]
        except KeyError:
            shutterstate = 1  # other types are comps, so open shutter

        if shutterstate:
            azcam.db.tools["controller"].set_int_ms(int(self.exposure_time * 1000))
            # azcam.db.tools["controller"].set_no_int_ms(0)
            azcam.db.tools["controller"].set_no_int_ms(self.shutter_delay)
        else:
            azcam.db.tools["controller"].set_no_int_ms(int(self.exposure_time * 1000))
            azcam.db.tools["controller"].set_int_ms(0)

        return

    def get_exposuretime(self):
        """
        Get current exposure time in seconds.
        """

        # not read directly from Archon controller
        return self.exposure_time

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read.
        """

        return azcam.db.tools["controller"].get_pixels_remaining()


class ArchonFileConverter(object):
    """
    Class to convert Archon image file.
    """

    def __init__(self):
        # save MEF file
        self.save_file = 0
        # output buffer
        self.o_data = 0

        # data types for fits images
        self.data_types = {
            8: "uint8",
            16: "uint16",
            32: "uint32",
            64: "uint64",
            -32: "float32",
            -64: "float64",
        }

        self.detector_filename = ""

        self.overwrite_infile = 0
        self.overwrite_outfile = 0

        self.fSize = 0
        self.out_file = ""
        self.curr_file = ""

        self.hdulist = ""

        # values read from the configuration file
        self.detname = ""
        self.det_number = []
        self.detpos_x = []
        self.detpos_y = []
        self.extpos_x = []
        self.extpos_y = []
        self.detformat = []
        self.focalplane = []
        self.roi = []
        self.amppix1 = []
        self.amppix2 = []
        self.ext_name = []
        self.ext_number = []
        self.jpg_ext = []
        self.amp_cfg = []

        self.ns_total = 0
        self.ns_predark = 0
        self.ns_underscan = 0
        self.ns_overscan = 0
        self.np_total = 0
        self.np_predark = 0
        self.np_underscan = 0
        self.np_overscan = 0
        self.np_frametransfer = 0

        self.is_valid = 1
        self.asmsize = (0, 0)
        self.size_x = 0
        self.size_y = 0
        self.from_file = 0

        # self.num_extensions = 16
        # self.offsets = 16 * [0.0]
        # self.scales = 16 * [1.0]

        self.prescan1 = 0
        self.prescan2 = 0

        self.exptime = 0.0
        self.intms = 0
        self.nointms = 0

    def copy_to_buffer(self, in_buffer, out_buffer):
        """
        Copies the data from the controller to a data buffer.
        """

        self.InData = in_buffer
        self.o_data = out_buffer

        self.buffer_processing()

        return

    def buffer_processing(self):
        """
        Process input buffer to output MEF file.
        """

        frameBase = "BUF%d" % (azcam.db.tools["controller"].read_buffer)

        self.NAMPS = self.numparamps * self.numseramps
        self.NAXIS1 = int(azcam.db.tools["controller"].dict_frame[frameBase + "WIDTH"])
        self.NAXIS2 = int(azcam.db.tools["controller"].dict_frame[frameBase + "HEIGHT"])
        self.PIXELS = int(azcam.db.tools["controller"].dict_frame[frameBase + "PIXELS"])
        self.LINES = int(azcam.db.tools["controller"].dict_frame[frameBase + "LINES"])

        self.data_type = "<u2"
        self.exptime = float(azcam.db.tools["controller"].int_ms / 1000)
        self.intms = azcam.db.tools["controller"].int_ms
        self.nointms = azcam.db.tools["controller"].noint_ms

        # make a 1D copy on the input data
        self.data = numpy.ndarray(
            shape=(self.NAXIS1 * self.NAXIS2), buffer=self.InData, dtype=self.data_type
        ).copy()

        # current line of self.PIXELS pixels
        cntLine = 0

        self.StartTime = time.time()

        if azcam.db.tools["exposure"].image.focalplane.num_detectors > 1:
            # reshape the input data array copy
            # self.NData = self.data.reshape(self.NAMPS * self.LINES, self.PIXELS).copy()
            self.NData = self.data.reshape(
                [self.NAMPS, self.PIXELS * self.LINES]
            ).copy()

            # loop through all lines in Archon buffer
            for currLine in range(0, self.LINES):
                for posY in range(0, self.numparamps):
                    for posX in range(0, self.numseramps):
                        posAmp = posX + posY * self.numseramps

                        # calc amp position in actual mosaic
                        indxAmp = (
                            (self.extpos_y[posAmp] - 1) * self.numseramps
                            + self.extpos_x[posAmp]
                            - 1
                        )

                        startpos = currLine * self.NAXIS1 + posAmp * self.PIXELS
                        endpos = startpos + self.PIXELS
                        dataline = self.data[startpos:endpos]

                        # break data into amp sections, no flipping
                        self.o_data[indxAmp][
                            currLine * self.PIXELS : currLine * self.PIXELS
                            + self.PIXELS
                        ] = dataline

                        # if self.amp_cfg[posAmp] == 0:
                        #     # no flip
                        #     self.o_data[indxAmp][
                        #         currLine * self.PIXELS : currLine * self.PIXELS
                        #         + self.PIXELS
                        #     ] = dataline

                        # elif self.amp_cfg[posAmp] == 1:
                        #     # flip X
                        #     self.o_data[indxAmp][
                        #         currLine * self.PIXELS : currLine * self.PIXELS
                        #         + self.PIXELS
                        #     ] = dataline[::-1]

                        # elif self.amp_cfg[posAmp] == 2:
                        #     # flip Y
                        #     self.o_data[indxAmp][
                        #         # (self.LINES - currLine - 1)
                        #         # * self.PIXELS : (self.LINES - currLine)
                        #         # * self.PIXELS
                        #         (self.LINES - currLine - 1)
                        #         * self.PIXELS : (self.LINES - currLine - 1)
                        #         * self.PIXELS
                        #         + self.PIXELS
                        #     ] = dataline
                        # else:
                        #     # flip XY
                        #     self.o_data[indxAmp][
                        #         # (self.LINES - currLine - 1)
                        #         # * self.PIXELS : (self.LINES - currLine)
                        #         # * self.PIXELS
                        #         (self.LINES - currLine - 1)
                        #         * self.PIXELS : (self.LINES - currLine - 1)
                        #         * self.PIXELS
                        #         + self.PIXELS
                        #     ] = dataline[::-1]

            # only for new 90prime  19nov24
            # nx = int(self.NAXIS1 / self.numseramps)
            # ny = int(self.NAXIS2 / self.numparamps)
            # for ampnum in range(4):
            # if self.amp_cfg[ampnum] == 0:
            #     pass
            # elif self.amp_cfg[ampnum] == 1:
            #     buff2d = self.o_data[ampnum].reshape(ny, nx)
            #     buff2d_flipped = numpy.flip(buff2d, 1)
            #     self.o_data[ampnum] = buff2d_flipped.reshape(nx * ny)
            # elif self.amp_cfg[ampnum] == 2:
            #     buff2d = self.o_data[ampnum].reshape(ny, nx)
            #     buff2d_flipped = numpy.flip(buff2d, 0)
            #     self.o_data[ampnum] = buff2d_flipped.reshape(nx * ny)
            # elif self.amp_cfg[ampnum] == 3:
            #     buff2d = self.o_data[ampnum].reshape(ny, nx)
            #     buff2d_flipped = numpy.flip(buff2d, 1)
            #     buff2d_flipped = numpy.flip(buff2d_flipped, 0)
            #     self.o_data[ampnum] = buff2d_flipped.reshape(nx * ny)

            # buff2d = self.o_data[ampnum].reshape(ny, nx)
            # buff2d_flipped = numpy.flip(buff2d, 0)
            # self.o_data[ampnum] = buff2d_flipped.reshape(nx * ny)
            # pass

        else:
            # single CCD

            # reshape the input data array copy
            self.NData = self.data.reshape(self.NAMPS * self.LINES, self.PIXELS).copy()

            for posY in range(0, self.numparamps):
                currPart = posY * self.numseramps
                for currLine in range(0, self.LINES):
                    for posX in range(0, self.numseramps):
                        posAmp = posX + currPart
                        indxAmp = (
                            (self.extpos_y[posAmp] - 1) * self.numseramps
                            + self.extpos_x[posAmp]
                            - 1
                        )
                        if self.amp_cfg[posAmp] == 0:
                            # no flip
                            self.o_data[indxAmp][
                                currLine * self.PIXELS : (currLine + 1) * self.PIXELS
                            ] = self.NData[cntLine]
                        elif self.amp_cfg[posAmp] == 1:
                            # flip X
                            self.o_data[indxAmp][
                                currLine * self.PIXELS : (currLine + 1) * self.PIXELS
                            ] = self.NData[cntLine][::-1]
                        elif self.amp_cfg[posAmp] == 2:
                            # flip Y
                            self.o_data[indxAmp][
                                (self.LINES - currLine - 1)
                                * self.PIXELS : (self.LINES - currLine)
                                * self.PIXELS
                            ] = self.NData[cntLine]
                        else:
                            # flip XY
                            self.o_data[indxAmp][
                                (self.LINES - currLine - 1)
                                * self.PIXELS : (self.LINES - currLine)
                                * self.PIXELS
                            ] = self.NData[cntLine][::-1]
                        cntLine += 1

        self.StopTime = time.time()

        return

    def set_detector_config(self, sensor_data):
        """
        Set detector configuration parameters.
        """

        """
        Notes:
            # ExtPosN - in ds9 - (1,1) is bottem left
            # ExtNum for IMAGEID
        """

        # defaults may have been set already
        if sensor_data.get("name"):
            self.detname = sensor_data["name"]
        if sensor_data.get("det_number"):
            self.det_number = sensor_data["det_number"]
        if sensor_data.get("format"):
            self.detformat = sensor_data["format"]
        if sensor_data.get("focalplane"):
            self.focalplane = sensor_data["focalplane"]
        if sensor_data.get("roi"):
            self.roi = sensor_data["roi"]
        if sensor_data.get("ext_name"):
            self.ext_name = sensor_data["ext_name"]
        if sensor_data.get("ext_number"):
            self.ext_number = sensor_data["ext_number"]
        if sensor_data.get("jpg_order"):
            self.jpg_ext = sensor_data["jpg_order"]

        # maybe fix this
        self.amp_cfg = self.focalplane[4]

        self.extpos_x = [x[0] for x in sensor_data["ext_position"]]
        self.extpos_y = [x[1] for x in sensor_data["ext_position"]]
        self.detpos_x = [x[0] for x in sensor_data["det_position"]]
        self.detpos_y = [x[1] for x in sensor_data["det_position"]]
        # self.amppix1 = [x[0] for x in sensor_data["amp_pixel_position"]]
        # self.amppix2 = [x[1] for x in sensor_data["amp_pixel_position"]]

        self.ns_total = self.detformat[0]
        self.ns_predark = self.detformat[1]
        self.ns_underscan = self.detformat[2]
        self.ns_overscan = self.detformat[3]
        self.np_total = self.detformat[4]
        self.np_predark = self.detformat[5]
        self.np_underscan = self.detformat[6]
        self.np_overscan = self.detformat[7]
        self.np_frametransfer = self.detformat[8]

        self.numseramps = self.focalplane[2]
        self.numparamps = self.focalplane[3]

        return


class ReceiveDataArchon(object):
    """
    Exposure subclass to receive image data.
    """

    def __init__(self, exposure):
        self.exposure = exposure  # upper level exposure object

        self.DeinterlaceParams = ""
        self.port = 0
        self.is_valid = 1

        self.PixelsReadout = 0
        self.pixels_remaining = 0
        self.pixels_remaining = 0
        self.camserver = 0
        self.socket = 0

        # Deinterlace mode
        self.deinterlace_mode = 1
        # Number of pixels per amplifier
        self.numpix_amp = 0
        # Number of amplifiers
        self.numamps_image = 0

    def receive_archon_image_data(self):
        """
        Receives image data (and raw data if rawdata_enable=1) from the Archon controller.
        Data goes to controller.imagedata as one buffer.
        """

        # initial values
        self.PixelsReadout = 0
        self.pixels_remaining = 0

        # values taken from the Archon GUI
        BURST_LEN = 1024
        lineSize = BURST_LEN
        # chunkSize = 1024 * BURST_LEN
        rawBlockSize = 2048

        if (
            azcam.db.tools["controller"].read_buffer > 0
            and azcam.db.tools["controller"].read_buffer < 4
        ):
            frameBase = "BUF%d" % (azcam.db.tools["controller"].read_buffer)
            frame = frameBase + "FRAME"

            if int(azcam.db.tools["controller"].dict_frame[frame]) > 0:
                # frame buffer base address
                addr = int(azcam.db.tools["controller"].dict_frame[frameBase + "BASE"])
                # get frame width and height
                frameW = int(
                    azcam.db.tools["controller"].dict_frame[frameBase + "WIDTH"]
                )
                frameH = int(
                    azcam.db.tools["controller"].dict_frame[frameBase + "HEIGHT"]
                )
                # get sample mode
                sampleMode = (
                    int(azcam.db.tools["controller"].dict_frame[frameBase + "SAMPLE"])
                    + 1
                )

                # calculate fetch command values
                frameSize = sampleMode * 2 * frameW * frameH
                lines = int((frameSize + lineSize - 1) / lineSize)
                rawBlocks = int(
                    azcam.db.tools["controller"].dict_frame[frameBase + "RAWBLOCKS"]
                )
                rawLines = int(
                    azcam.db.tools["controller"].dict_frame[frameBase + "RAWLINES"]
                )
                rawSize = rawBlocks * rawLines * rawBlockSize
                rawOffset = int(
                    azcam.db.tools["controller"].dict_frame[frameBase + "RAWOFFSET"]
                )

                cmd = "FETCH%08X%08X" % (addr, lines)

                azcam.db.tools["controller"].archon_bin_command(cmd)

                # frameSize = frameSize + 4
                totalRecv = 0

                self.TData = numpy.empty(shape=int(frameSize / 2), dtype="<u2")

                self.pixels_remaining = frameSize / 2
                self.PixelsReadout = 0

                # total bytes received from the Archon controller | each data frame starts with preamble '<NN:' | last data frame includes padding bytes
                totalBytes = lines * (lineSize + 4)

                totalRecv = 0
                currLine = lineSize + 4

                dataBuff = b""
                totalPix = 0
                imgPix = frameSize / 2

                while totalRecv < totalBytes:
                    getData = azcam.db.tools["controller"].camserver.socket.recv(
                        currLine
                    )

                    totalRecv += len(getData)

                    dataBuff += getData
                    currLen = len(dataBuff)

                    if currLen >= 1028:
                        # got one fulle data frame
                        if dataBuff[3] == 58:
                            if totalPix + 512 < imgPix:
                                pixCnt = 512
                            else:
                                pixCnt = imgPix - totalPix

                            currBytes = 2 * pixCnt

                            lData = int(currBytes + 4)
                            ImageBufferTemp = numpy.ndarray(
                                shape=(int(pixCnt)),
                                dtype="<u2",
                                buffer=dataBuff[4:lData],
                            )

                            lData = int(totalPix + pixCnt)
                            self.TData[totalPix:lData] = ImageBufferTemp[
                                0 : int(pixCnt)
                            ]
                            totalPix += pixCnt

                            dataBuff = dataBuff[1028:]
                        else:
                            azcam.log("-----> dataBuff length:", len(dataBuff), level=3)
                            totalRecv = totalBytes

                azcam.log("----> totalRecv =", totalRecv, level=3)
                if totalRecv == totalBytes:
                    self.exposure.image.valid = 1
                    self.pixels_remaining = 0
                    azcam.db.tools["controller"].imagedata = self.TData

                    if azcam.db.tools["controller"].rawdata_enable == 1:
                        # receive rad data
                        lines = int((rawSize + lineSize - 1) / lineSize)

                        cmd = "FETCH%08X%08X" % (addr + rawOffset, lines)

                        azcam.db.tools["controller"].archon_bin_command(cmd)

                        # frameSize = frameSize + 4
                        totalRecv = 0

                        self.RData = numpy.empty(shape=int(rawSize / 2), dtype="<u2")

                        self.pixels_remaining = rawSize / 2
                        self.PixelsReadout = 0

                        # total bytes received from the Archon controller | each data frame starts with preamble '<NN:' | last data frame includes padding bytes
                        totalBytes = lines * (lineSize + 4)

                        totalRecv = 0
                        currLine = lineSize + 4

                        dataBuff = b""
                        totalPix = 0
                        imgPix = frameSize / 2

                        # total bytes received from the Archon controller | each data frame starts with preamble '<NN:' | last data frame includes padding bytes
                        totalBytes = lines * (lineSize + 4)

                        totalRecv = 0

                        self.RData = numpy.empty(shape=int(rawSize / 2), dtype="<u2")

                        while totalRecv < totalBytes:
                            getData = azcam.db.tools[
                                "controller"
                            ].camserver.socket.recv(currLine)

                            totalRecv += len(getData)

                            dataBuff += getData
                            currLen = len(dataBuff)

                            if currLen >= 1028:
                                # got one fulle data frame
                                if dataBuff[3] == 58:
                                    if totalPix + 512 < imgPix:
                                        pixCnt = 512
                                    else:
                                        pixCnt = imgPix - totalPix

                                    currBytes = 2 * pixCnt

                                    lData = int(currBytes + 4)
                                    ImageBufferTemp = numpy.ndarray(
                                        shape=(int(pixCnt)),
                                        dtype="<u2",
                                        buffer=dataBuff[4:lData],
                                    )

                                    lData = int(totalPix + pixCnt)
                                    self.RData[totalPix:lData] = ImageBufferTemp[
                                        0 : int(pixCnt)
                                    ]
                                    totalPix += pixCnt

                                    dataBuff = dataBuff[1028:]
                                else:
                                    azcam.log(
                                        "-----> dataBuff length:",
                                        len(dataBuff),
                                        level=3,
                                    )
                                    azcam.log("ERROR ", dataBuff[:4].decode(), level=3)
                                    totalRecv = totalBytes

                        if totalRecv == totalBytes:
                            azcam.db.tools["controller"].rawdata = self.RData

                    return

                else:
                    if self.exposure_flag != self.EF_ABORT:
                        s = "ERROR did not receive entire image buffer"
                        raise azcam.exceptions.AzcamError(s)
                    else:
                        raise azcam.exceptions.AzcamError("Exposure ABORTED")

            else:
                raise azcam.exceptions.AzcamError("No frame available for fetching")

        else:
            raise azcam.exceptions.AzcamError("Wrong frame number")



"""
Contains the Focus class.
"""

import time

import azcam
import azcam.utils
import azcam.exceptions
from azcam.tools.tools import Tools


class Focus(Tools):
    """
    Class for focusing a camera.

    Either the telescope or instrument may be moved for focus adjustment.
    The focus sequence performed is:

    - integrate
    - move focus
    - shift detector (2x last time)
    - (repeat above steps)
    - readout
    - return to starting focus position
    - save image
    """

    def __init__(self, tool_id="focus", description=None):
        """
        Create focus tool.
        """

        super().__init__(tool_id, description)

        #: Number of exposures in focus sequence
        self.number_exposures = 7
        #: Number of focus steps between each exposure in a frame
        self.focus_step = 30
        #: Number of rows to shift detector for each focus step
        self.detector_shift = 10
        #: current focus position
        self.focus_position = 0
        #: exposure time
        self.exposure_time = 1.0
        #: focus component for motion - instrument or telescope
        self.focus_component = "instrument"
        #: focus type, absolute or step
        self.focus_type = "absolute"
        #: flag to not prompt when set_pars already called
        self.set_pars_called = 0
        #: delay in seconds between exposures
        self.move_delay = 3

    def initialize(self):
        """
        Initialize focus routine.
        """

        self.exposure = azcam.db.tools["exposure"]
        try:
            self.instrument = azcam.db.tools["instrument"]
        except Exception:
            self.instrument = None
        try:
            self.telescope = azcam.db.tools["telescope"]
        except Exception:
            self.telescope = None
            pass

        self.is_initialized = 1

        return

    def abort(self):
        """
        Abort focus exposure.
        """

        self.exposure.abort()

        return

    def set_pars(
        self,
        exposure_time: float,
        number_exposures: int = 7,
        focus_step: float = 30,
        detector_shift: int = 10,
    ):
        """
        Set focus related parameters.
        Args:
            number_exposures: Number of exposures in focus sequence.
            focus_step: Number of focus steps between each exposure in a frame.
            detector_shift: Number of rows to shift detector for each focus step.
            exposuretime: Exposure time i seconds.

        """

        self.exposure_time = float(exposure_time)
        self.number_exposures = int(number_exposures)
        self.focus_step = float(focus_step)
        self.detector_shift = int(detector_shift)

        self.set_pars_called = 1

        return

    def _get_focus(self) -> float:
        """
        Get focus using default focus_id.
        """

        if self.focus_component == "instrument":
            reply = self.instrument.get_focus()
        elif self.focus_component == "telescope":
            reply = self.telescope.get_focus()

        return float(reply)

    def _set_focus(self, focus_value: float):
        """
        Set focus using default focus_id and focus_type
        """

        if self.focus_component == "instrument":
            return self.instrument.set_focus(focus_value, focus_type=self.focus_type)
        elif self.focus_component == "telescope":
            return self.telescope.set_focus(focus_value, focus_type=self.focus_type)

    def save_keywords(self):
        """
        Save focus related keywords for image header.
        """

        if self.focus_component == "instrument":
            reply = self.instrument.get_focus()
            self.instrument.set_keyword(
                "FOCSTART", reply, "starting focus value", "str"
            )
            self.instrument.set_keyword(
                "FOCSTEP", self.focus_step, "focus step size", "float"
            )
            self.instrument.set_keyword(
                "FOCSHIFT", self.detector_shift, "focus detector shift pixels", "int"
            )
            self.instrument.set_keyword(
                "FOCSTEPS", self.number_exposures, "focus number of exposures", "int"
            )
        elif self.focus_component == "telescope":
            reply = self.telescope.get_focus()
            self.telescope.set_keyword("FOCSTART", reply, "starting focus value", "str")
            self.telescope.set_keyword(
                "FOCSTEP", self.focus_step, "focus step size", "float"
            )
            self.telescope.set_keyword(
                "FOCSHIFT", self.detector_shift, "focus detector shift pixels", "int"
            )
            self.telescope.set_keyword(
                "FOCSTEPS", self.number_exposures, "focus number of exposures", "int"
            )

        return

    def run(
        self,
        exposure_time: float = 1.0,
        number_exposures: int = 7,
        focus_step: float = 30,
        detector_shift: int = 10,
    ):
        """
        Execute the focus sequence.
        If focus.set_pars() was previously called then those values are used and input here is ignored.
        Args:
            number_exposures: Number of exposures in focus sequence.
            focus_step: Number of focus steps between each exposure in a frame.
            detector_shift: Number of rows to shift detector for each focus step.
            exposuretime: Exposure time in seconds.
        """

        if self.set_pars_called:
            pass

        else:
            self.exposure_time = float(exposure_time)
            self.number_exposures = int(number_exposures)
            self.focus_step = int(focus_step)
            self.detector_shift = int(detector_shift)

        abort_flag = 0

        # exposure time - zero not allowed for focus
        self.exposure.set_exposuretime(self.exposure_time)
        ExpTime = self.exposure.get_exposuretime()
        if ExpTime < 0.001:
            azcam.exceptions.warning("do not focus with zero exposure time")
            return
        self.exposure.set_exposuretime(self.exposure_time)

        # save parameters to be changed
        root = azcam.db.parameters.get_par("imageroot")
        includesequencenumber = azcam.db.parameters.get_par(
            "imageincludesequencenumber"
        )
        autoname = azcam.db.parameters.get_par("imageautoname")
        autoincrementsequencenumber = azcam.db.parameters.get_par(
            "imageautoincrementsequencenumber"
        )
        title = azcam.db.parameters.get_par("imagetitle")
        testimage = azcam.db.parameters.get_par("imagetest")
        imagetype = azcam.db.parameters.get_par("imagetype")

        azcam.db.parameters.set_par("imageroot", "focus.")
        azcam.db.parameters.set_par("imageincludesequencenumber", 1)
        azcam.db.parameters.set_par("imageautoname", 0)
        azcam.db.parameters.set_par("imageautoincrementsequencenumber", 1)
        azcam.db.parameters.set_par("imagetest", 0)
        azcam.db.parameters.set_par("imageoverwrite", 1)

        # set header keywords
        self.save_keywords()

        # start
        self.exposure.begin(self.exposure_time, "object", "Focus")

        # loop over FocusNumber integrations
        current_exposure = 1

        # get starting focus
        current_focus_position = self._get_focus()
        starting_focus_value = current_focus_position

        nsteps = 0  # total number of focus steps
        while current_exposure <= self.number_exposures:
            # check for abort
            k = azcam.utils.check_keyboard(0)
            ab = azcam.db.abortflag
            if k == "q" or ab:
                abort_flag = 1
                break

            if current_exposure > 1:
                if self.focus_type == "step":
                    self._set_focus(self.focus_step)
                    nsteps += self.focus_step
                elif self.focus_type == "absolute":
                    self._set_focus(
                        current_focus_position + self.focus_step,
                    )
                time.sleep(self.move_delay)
                current_focus_position = self._get_focus()

                # shift detector
                self.exposure.parshift(self.detector_shift)
                if current_exposure == self.number_exposures:
                    azcam.log("Last exposure, double shifting")
                    self.exposure.parshift(self.detector_shift)

            azcam.log(
                f"Next exposure is {current_exposure} of {self.number_exposures} at focus position {current_focus_position:.3f}"
            )

            # integrate
            azcam.log("Integrating")
            try:
                self.exposure.integrate()
            except azcam.exceptions.AzcamError:
                azcam.log("Focus exposure aborted")
                self._set_focus(starting_focus_value)
                azcam.db.parameters.set_par("imageroot", root)
                azcam.db.parameters.set_par(
                    "imageincludesequencenumber", includesequencenumber
                )
                azcam.db.parameters.set_par("imageautoname", autoname)
                azcam.db.parameters.set_par(
                    "imageautoincrementsequencenumber", autoincrementsequencenumber
                )
                azcam.db.parameters.set_par("imagetest", testimage)
                azcam.db.parameters.set_par("imagetitle", title)
                azcam.db.parameters.set_par("imagetype", imagetype)
                fp = self._get_focus()
                azcam.log(f"Current focus: {fp:.3f}")
                return

            # increment focus number
            current_exposure += 1

        # set focus back to starting position
        azcam.log(f"Returning focus to starting value {starting_focus_value:.3f}")
        if self.focus_type == "step":
            steps = -1 * nsteps
            self._set_focus(steps)
        elif self.focus_type == "absolute":
            self._set_focus(starting_focus_value)
        time.sleep(self.move_delay)
        fp = self._get_focus()
        azcam.log(f"Current focus: {fp:.3f}")

        if not abort_flag:
            # readout and finish
            azcam.log("Reading out")
            self.exposure.readout()
            self.exposure.end()
        else:
            azcam.db.parameters.set_par("ExposureFlag", azcam.db.exposureflags["NONE"])

        # finish
        azcam.db.parameters.set_par("imageroot", root)
        azcam.db.parameters.set_par("imageincludesequencenumber", includesequencenumber)
        azcam.db.parameters.set_par("imageautoname", autoname)
        azcam.db.parameters.set_par(
            "imageautoincrementsequencenumber", autoincrementsequencenumber
        )
        azcam.db.parameters.set_par("imagetest", testimage)
        azcam.db.parameters.set_par("imagetitle", title)
        azcam.db.parameters.set_par("imagetype", imagetype)

        return



import datetime
import os
import sys
import logging.handlers
import threading
import socket
from typing import List

import azcam
import azcam.exceptions
import loguru

import azcam.utils


class AzCamLogger(object):
    """
    The azcam Logger class.
    """

    def __init__(self) -> None:
        self.logfile = "azcam.log"
        self.logger = loguru.logger
        self.use_logprefix = 1
        self.last_data = []  # data since last call to get_data
        self.last_data_max = 1024  # max entries in last_data buffer

    def log(self, message: str, *args: List[str], prefix: str = "", level: int = 1):
        """
        Send a message to the logging system.

        Message is output to logger if level > db.verbosity.
        Levels are:
        0 => silent
        1 => normal
        2 => extended info
        3 => debug

        Args:
            message: String message to be logged
            args: Additional string message to be logged
            prefix: Prefix to be prepended to logged message, ex: 'Log-> '
            level: verbosity level for output

        """

        # don't log if level > global verbosity
        if level > azcam.db.verbosity:
            return

        message = str(message)  # better for exceptions
        message = azcam.utils.dequote(message)

        # format message
        if len(args) == 1:
            message = message + " " + str(args[0])
        elif len(args) > 1:
            message = message + " " + " ".join(str(x) for x in args)

        if prefix != "" and self.use_logprefix:
            message = prefix + message

        # log eveything at INFO level
        self.logger.info(f"{message}")

        # append to last_data
        self.last_data.append(f'"{message}"')

        return

    def _logfilter(self, record):
        # do not send messages from threads to console
        thread_name = threading.currentThread().getName()
        if thread_name == "MainThread":
            return 1
        else:
            return 1  # was 0

    def info(self, message: str, *args, **kwargs):
        return self.logger.info(message, args, kwargs)

    def warning(self, message: str, *args, **kwargs):
        return self.logger.warning(message, args, kwargs)

    def error(self, message: str, *args, **kwargs):
        return self.logger.error(message, args, kwargs)

    def start_logging(
        self, logtype="13", host="localhost", port=2404, logfile=None, use_timestamp=1
    ):
        """
        Start the azcam logger.

        :param logtype: code for loggers to start (1 console, 2 socket, 3 file - combine as '23')
        :param host: hostname for logging over socket
        :param port: socket port number
        :param logfile: base filename of log file. If not absolute path, will use db.systemfolder.
        :param use_timestamp: append timestamp to logfile name.
        """

        # remove default logger for customization
        try:
            self.logger.remove(0)
        except Exception:
            pass

        # console handler
        if "1" in logtype:
            self.logger.add(
                sys.stdout,
                level="INFO",
                format="{message}",
                filter=self._logfilter,
                colorize=True,
                enqueue=True,
                # backtrace=True,
                # diagnose=True,
            )

        # socket handler
        if "2" in logtype:
            socket_handler = logging.handlers.SocketHandler(
                "localhost",
                port,
            )
            self.logger.add(socket_handler)
            azcam.log(f"Logging to logging server on port {port}")

        # rotating file handler
        if "3" in logtype:
            if logfile is None:
                if self.logfile is None:
                    raise azcam.exceptions.AzcamError("no logfile specified")
            else:
                self.logfile = logfile
            if use_timestamp:
                tt = datetime.datetime.strftime(
                    datetime.datetime.now(), "%d%b%y_%H%M%S"
                )
                s1, s2 = os.path.splitext(self.logfile)
                self.logfile = f"{s1}_{tt}{s2}"

            self.logger.add(
                self.logfile,
                format="{time:DD-MMM-YY HH:mm:ss.SSS} | {level} | {message}",
                rotation="10 MB",
                retention="1 week",
            )
            azcam.log(f"Logging to file {self.logfile}")

        return

    def get_logdata(self):
        """
        Returns log data.
        """

        buffer = self.last_data
        self.last_data = []

        return buffer


def check_for_remote_logger(host: str = "localhost", port: int = 2404):
    """
    Check if a remote logging server is running.
    """

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)
    try:
        sock.settimeout(0.1)
        sock.connect((host, port))
        sock.close()
        return True
    except Exception:
        return False



"""
DASH web server for azcam
"""

import threading
import logging

from dash import Dash, html, dcc, callback, Input, Output, State, ctx
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam
from azcam.web.queue.queue_card import queue_card


class QueueWeb(object):
    """
    Web code for queue operations using Dash (plotly).
    """

    def __init__(self) -> None:

        azcam.db.queueserver = self

        self.app = Dash(
            __name__,
            external_stylesheets=[dbc.themes.BOOTSTRAP],
            suppress_callback_exceptions=False,
            requests_pathname_prefix="/queue/",
        )
        logging.getLogger("werkzeug").setLevel(logging.CRITICAL)  # stop messages

        self.queue_card = queue_card()

        # app layout
        self.app.layout = html.Div(
            [
                self.queue_card,
                dcc.Interval("queue_interval", interval=500, n_intervals=0),
            ]
        )

    # interval update
    @callback(
        Output("message_queue", "children"),
        Input("queue_interval", "n_intervals"),
    )
    def update_queue(n):
        """
        Update queue status.
        """

        # the return order must match Output order
        data = "mike"

        return [
            str(n),
        ]

        return

    def set_message(self, message: str = "") -> None:
        """
        Set the browser message.
        """

        return
